from sage.all import *

import common

import itertools

import functools

from util import create_logger, cached_simple_method

logger = create_logger(__name__)

from convex import conify_polyhedron, is_contained_in_dual, dual_cone_as_polyhedron, RationalSet
from convex import StrictlyPositiveOrthant, PositiveOrthant
from convex import inner_open_normal_fan, vertex_by_direction

from util import minimal_elements, principal_minors, split_off_torus, symbolic_to_ratfun

from laurent import LaurentIdeal, LaurentPolynomial

from tmplist import TemporaryList

from toric import is_nondegenerate

from torus import SubvarietyOfTorus

from abstract import ZetaDatum, ReductionError, TopologicalZetaProcessor, LocalZetaProcessor
from surf import SURF

from cycrat import CyclotomicRationalFunction

DEPTH_BOUND = 24

class Breakout(Exception):
    pass

class IgusaDatum(ZetaDatum):
    # General enough for representation and Igusa-type zeta functions.
    def __init__(self, ideals, depth=0):
        self.ideals = list(ideals)
        if not self.ideals:
            raise ValueError('need at least one ideal')

        self.ring = self.ideals[0].ring
        if any(J.ring != self.ring for J in self.ideals):
            raise ValueError('inconsistent base rings')
        
        self.ambient_dim = self.ring.ngens()
        self.RS = self.ideals[0].RS

        if any(I.RS != self.RS for I in self.ideals):
            raise ValueError('inconsistent rational sets')

        self.polynomials = []
        self.initials = []

        self._poly2ideal = []
        self._factors = []
        self._shifts = []
        self._ideal2poly = []

        self._depth = depth

        for i, I in enumerate(self.ideals): # I = self.ideals[i]
            self._ideal2poly.append([])
            self._factors.append([])
            self._shifts.append([])
            
            for j, f in enumerate(I.gens): # f == I.gens[j]
                k = next((k for k, g in enumerate(self.polynomials) if g.divides(f) and (f/g).is_term()), None)

                if k is None:
                    self._ideal2poly[i].append(len(self.polynomials))
                    self._factors[i].append(LaurentPolynomial(self.ring.one()))
                    self._shifts[i].append(vector(ZZ, self.ring.ngens()))

                    self.polynomials.append(f)
                    self.initials.append(I.initials[j])
                    self._poly2ideal.append([(i,j)])

                else: # f == term * self.polynomials[k]
                    self._ideal2poly[i].append(k)
                    term = f/self.polynomials[k]
                    self._factors[i].append(term)
                    self._shifts[i].append(vector(ZZ, term.exponents()[0]))

                    if I.initials[j] is None and self.initials[k] is not None:
                        I.initials[j] = term * self.initials[k]

                    if I.initials[j] is not None and self.initials[k] is None:
                        self.initials[k] = I.initials[j]/term

                        for a,b in self._poly2ideal[k]:
                            assert self.ideals[a].gens[b] == self.polynomials[k] * self._factors[a][b]
                            self.ideals[a].initials[b] = self.initials[k] * self._factors[a][b]

                    self._poly2ideal[k].append((i,j))

        # BEGIN SANITY CHECKS
        for i in xrange(len(self.ideals)):
            for j, f in enumerate(self.ideals[i].gens):
                k = self._ideal2poly[i][j]
                assert f == self.polynomials[k] * self._factors[i][j]
                if (self.ideals[i].initials[j] is not None) or (self.initials[k] is not None):
                    assert self.ideals[i].initials[j] == self.initials[k] * self._factors[i][j]

        for k, f in enumerate(self.polynomials):
            for i, j in self._poly2ideal[k]:
                assert self.ideals[i].gens[j] == f * self._factors[i][j]
                if (self.ideals[i].initials[j] is not None) or (self.initials[k] is not None):
                    assert self.ideals[i].initials[j] == self.initials[k] * self._factors[i][j]
        # END SANITY CHECKS

    def is_empty(self):
        return self.RS.is_empty()

    def weight(self):
        return sum(len(f.exponents())-1 for f in (self.initials if self.is_balanced() else self.polynomials))

    def simplify(self):
        return IgusaDatum([I.simplify() for I in self.ideals], depth=self._depth)

    def is_balanced(self):
        return not any(f is None for f in self.initials)

    def _replace_single_ideal(self, i, new, depth=None):
        # Create a new Igusa datum with RS == new.RS and i-th ideal new.
        # new.RS should better be contained in each self.ideals[j].RS;
        # otherwise initial forms should be deleted.
        return IgusaDatum([new if i == j
                           else LaurentIdeal(gens=J.gens, initials=J.initials,
                                             RS=new.RS, ring=J.ring, normalise=True)
                           for j,J in enumerate(self.ideals)],
                          depth=self._depth if depth is None else depth)

    def balance(self):
        if self.is_balanced():
            yield self
            return
        i,I = next((i,I) for i,I in enumerate(self.ideals) if not I.is_balanced())
        for J in I.balance():
            yield self._replace_single_ideal(i, J)

    @cached_simple_method
    def is_regular(self):
        if not self.is_balanced():
            return False
        return is_nondegenerate([f.num for f in self.initials],
                                all_subsets=True, all_initial_forms=False,
                                collision_handler=lambda J: setattr(self, '_coll_idx', J))

    # def is_ordered(self):
    #     return len(self.RS.polyhedra) <= 1

    # def order(self):
    #     for P in self.RS.polyhedra:
    #         RS = RationalSet(P)
    #         yield IgusaDatum([ LaurentIdeal(gens=I.gens, initials=I.initials,
    #                                         RS=RS, ring=I.ring, normalise=True)
    #                            for I in self.ideals], depth=self._depth)

    def _low_level_reduce(self, j, a, b, ta, tb, strict):
        return self._replace_single_ideal(j, self.ideals[j].reduce(a, b, ta, tb, strict))

    def _judge_candidate(self, j, a, b, ta, tb):
        # We us the exact same approach as in subobjects.greedy_chopper.
        lt = self._low_level_reduce(j, a, b, ta, tb, strict=False).simplify()
        rt = self._low_level_reduce(j, b, a, tb, ta, strict=True).simplify()
        children = [t.simplify() for t in itertools.chain(lt.balance(), rt.balance())]
        naughty = [t for t in children if (not t.is_regular()) and t.weight() >= self.weight()]
        for t in naughty:
            t._depth += 1
            logger.info('Depth increased to %d' % t._depth)
            if t._depth > DEPTH_BOUND:
                logger.info('Exceeded the given bound for the depth of reductions.')
                return Infinity, children

        if not naughty:
            return float(0), children

        return float(sum(t.weight() for t in naughty))/ (self.weight() * len(naughty)), \
            naughty + [c for c in children if c not in naughty]

    def reduce(self, preemptive):
        if preemptive:
            raise ReductionError

        assert not self.is_regular()
        if self._depth > DEPTH_BOUND:
            raise ReductionError('Exceeded bound for reduction depth.')

        if len(self._coll_idx) == 1:
            raise ReductionError('toric singularities')

        # Find j s.t. self.ideals[j] is generated by at least two elements
        # of self.polynomials indexed by self._coll_idx.
        # 'bad' then contains at lesat 2 indices w.r.t. self.polynomials corresponding
        # to self._coll_idx

        j, bad = next(
            ( (j,list(bad)) for (j, bad) in
              ((j, Set(self._ideal2poly[j]).intersection(self._coll_idx)) for j in xrange(len(self.ideals)))
              if len(bad) >= 2),
            (None, None) )

        # Q: If the collision is spread across different ideals,
        #    should we try to find an internal simplification within
        #    one of them that gets rid of it? 
        if j is None:
            raise ReductionError('Collision involves separate ideals')

        optval, optsol = Infinity, None

        # 'bad' from above consists of indices for self.polynomials.
        # Turn it into indices for self.ideals[j].gens
        bad = [next(b for a,b in self._poly2ideal[k] if a == j) for k in bad]

        try:
            for a,b in itertools.combinations(bad, 2):
                for ta,tb in itertools.product(self.ideals[j].initials[a].terms(),
                                               self.ideals[j].initials[b].terms()):
                    val, sol = self._judge_candidate(j, a, b, ta, tb)
                    if val < optval:
                        optval, optsol = val, sol
                    if not val:
                        raise Breakout
        except Breakout:
            pass

        if optsol is None:
            raise ReductionError('reduction failed')

        logger.info('Optimal value for reduction step: %s' % optval)
        for t in optsol:
            yield t

    def __repr__(self):
        s = 'Igusa datum of weight %d:\n' % self.weight()
        s += 'Polynomials: ' + ', '.join(str(f) for f in self.polynomials) + '\n'
        s += 'Initials: ' + ', '.join(str(f) for f in self.initials) + '\n'
        s += 'Ideals:\n'
        for i, I in enumerate(self.ideals):
            s += ('[%2d] ' % i) + str(I) + '\n'
        s += repr(self.RS)
        return s

def _sqrt(f):
    if not f:
        return f
    li = list(f.factor())
    if any(e % 2 for _, e in li):
        raise ValueError('not a square')
    return f.parent(prod(a**(e//2) for (a,e) in li))

def evaluate_monomial_integral(mode, RS, polytopes, substitution, dims=None):
    # NOTE:
    # The name is slightly misleading since the factor (1-1/q)**RS.ambient_dim
    # is missing in the p-adic case.

    for P in polytopes:
        if P.ambient_dim() != RS.ambient_dim:
            raise RuntimeError('dimension mismatch')

    if not polytopes or any(P.is_empty() for P in polytopes):
        raise ValueError('need a non-empty collection of non-empty polytopes')

    for polyhedron in inner_open_normal_fan(sum(polytopes)):
        y = vector(polyhedron.vertices()[0])
        alphas = [vertex_by_direction(Q, y) for Q in polytopes]

        local_RS = RS & RationalSet(polyhedron)
        if local_RS.is_empty():
            continue

        A = matrix(ZZ, alphas)
        Phi = matrix(ZZ, [vector(ZZ, substitution[0]) * A,
                          vector(ZZ, substitution[1]) * A - vector(ZZ, A.ncols() * [1])]).transpose()

        if mode == 'topological':
            if local_RS.dim() < RS.dim():
                continue
            for surf in local_RS.topologise(Phi, dims=dims):
                yield surf
        elif mode == 'p-adic':
            with TemporaryList() as tmp_list:
                sm = local_RS.generating_function(base_list=tmp_list)
                for z in sm.monomial_substitution(QQ['t','q'], Phi):
                    yield z
        else:
            raise ValueError('unknown mode')
            
topologically_evaluate_monomial_integral = functools.partial(evaluate_monomial_integral, 'topological')
padically_evaluate_monomial_integral = functools.partial(evaluate_monomial_integral, 'p-adic', dims=None)

class IgusaProcessor(TopologicalZetaProcessor, LocalZetaProcessor):
    def __init__(self, *polynomials):
        if not polynomials:
            raise ValueError('need at least one polynomial')

        polynomials = [f if isinstance(f, LaurentPolynomial) else LaurentPolynomial(FractionField(f.parent())(f)) for f in polynomials]
        self.nvars = polynomials[0].nvars
        self.ideal = LaurentIdeal(gens=polynomials, RS=RationalSet(PositiveOrthant(self.nvars)), normalise=True)
    
    def root(self):
        return IgusaDatum([self.ideal])

    def topologically_evaluate_regular(self, datum):
        if not datum.is_regular():
            raise ValueError('need a regular datum')

        euler_cap = {}
        torus_factor_dim = {}

        N = Set(range(len(datum.polynomials)))
        _,d,_ = split_off_torus([datum.initials[i].num for i in N])
        min_dim = datum.ambient_dim - d
        
        if datum.RS.dim() < min_dim:
            logger.debug('Totally irrelevant datum')
            return
            yield

        for I in Subsets(N):
            F = [datum.initials[i].num for i in I]
            V = SubvarietyOfTorus(F, torus_dim=datum.ambient_dim)
            U,W = V.split_off_torus()
            torus_factor_dim[I] = W.torus_dim
            euler_cap[I] = U.khovanskii_characteristic() if U.is_nondegenerate() else U.euler_characteristic()
            assert torus_factor_dim[I] >= min_dim

        for I in Subsets(N):
            chi = sum((-1)**len(J) * euler_cap[I+J] for J in Subsets(N-I) if torus_factor_dim[I+J] == min_dim)
            if not chi:
                logger.debug('Vanishing Euler characteristic: I = %s' % I)
                continue
            
            I = list(I)
            polytopes = []

            id = identity_matrix(ZZ, len(I))
            def vectorise(k, vec):
                w = id[I.index(k)] if k in I else vector(ZZ,len(I))
                return vector(ZZ, list(vec) + list(w))

            assert len(datum._ideal2poly[0]) == len(datum.ideals[0].gens)
            polytopes = [Polyhedron(vertices=[ vectorise(k, datum.ideals[0].initials[m].exponents()[0]) for m,k in enumerate(datum._ideal2poly[0]) ], ambient_dim=datum.ambient_dim+len(I))]

            extended_RS = datum.RS * RationalSet(StrictlyPositiveOrthant(len(I)))

            assert all(extended_RS.ambient_dim == P.ambient_dim() for P in polytopes)
            for surf in topologically_evaluate_monomial_integral(extended_RS, polytopes, 
                                                                 [ (1,), (0,) ],
                                                                 dims=[min_dim + len(I)],
                                                                 ):
                yield SURF(scalar=chi*surf.scalar, rays=surf.rays)

    def padically_evaluate_regular(self, datum):
        if not datum.is_regular():
            raise ValueError('need a regular datum')

        q = var('q')
        count_cap = {}
        N = Set(range(len(datum.polynomials)))

        for I in Subsets(N):
            F = [datum.initials[i].num for i in I]
            V = SubvarietyOfTorus(F, torus_dim=datum.ambient_dim)
            count_cap[I] = V.count()

        for I in Subsets(N):
            cnt = sum((-1)**len(J) * count_cap[I+J] for J in Subsets(N-I))
            if not cnt:
                continue
            
            I = list(I)
            polytopes = []

            id = identity_matrix(ZZ, len(I))
            def vectorise(k, vec):
                w = id[I.index(k)] if k in I else vector(ZZ,len(I))
                return vector(ZZ, list(vec) + list(w))

            assert len(datum._ideal2poly[0]) == len(datum.ideals[0].gens)
            polytopes = [Polyhedron(vertices=[ vectorise(k, datum.ideals[0].initials[m].exponents()[0]) for m,k in enumerate(datum._ideal2poly[0]) ], ambient_dim=datum.ambient_dim+len(I))]

            extended_RS = datum.RS * RationalSet(StrictlyPositiveOrthant(len(I)))

            foo, ring = symbolic_to_ratfun(cnt * (q-1)**len(I)/q**datum.ambient_dim, [var('t'),var('q')])
            corr_cnt = CyclotomicRationalFunction.from_laurent_polynomial(foo, ring)

            assert all(extended_RS.ambient_dim == P.ambient_dim() for P in polytopes)

            for z in padically_evaluate_monomial_integral(extended_RS, polytopes, [ (1,), (0,) ]):
                yield corr_cnt * z

    def __repr__(self):
        return 'Igusa processor. IDEAL: %s' % self.ideal

class RepresentationProcessor(TopologicalZetaProcessor, LocalZetaProcessor):
    def __init__(self, arg):
        try:
            self.R, self.S = list(arg)
            self.algebra = None
        except TypeError:
            self.algebra = arg

    @cached_simple_method
    def root(self):
        if self.algebra is not None:
            self.R, self.S = self.algebra._commutator_matrices()
        self.ring = self.R.base_ring()

        d = self.ring.ngens()
        self.d = d

        polyhedra = []
        for i in xrange(d):
            eqns = [
                (i+1) * [0] + [1] + (d-i-1) * [0]
            ]
            ieqs = [
                [-1] + j * [0] + [1] + (d-j-1) * [0]
                for j in xrange(i)
            ]
            ieqs += [
                (j+1) * [0] + [1] + (d-j-1) * [0]
                for j in xrange(i+1,d)
            ]
            polyhedra.append(Polyhedron(eqns=eqns, ieqs=ieqs))
        self.RS = RationalSet(polyhedra, ambient_dim=d)

        F = FractionField(self.ring)
        two_u = matrix(F, self.R).rank() # self.R.rank()
        if two_u % 2:
            raise RuntimeError('this is odd')
        self.u = two_u // 2
        self.v = matrix(F, self.S).rank() # self.S.rank()

        if not d:
            return

        F = [
            LaurentIdeal(
                gens = [LaurentPolynomial(_sqrt(f)) for f in principal_minors(self.R, 2*j)],
                RS = self.RS,
                normalise = True)
            for j in range(0, self.u+1)
        ]

        G = [
            LaurentIdeal(
                gens = [LaurentPolynomial(g) for g in self.S.minors(j)],
                RS = self.RS,
                normalise = True)
            for j in range(0, self.v+1)
            ]

        oo = self.u + self.v + 2

        # On pairs:
        # The first component is used as is, the second is multiplied by the extra
        # variable. Note that index 0 corresponds to {1} and index oo to {0}.
        # We skip the |F_1|/|F_0 cap xF_1| factor which is generically trivial.
        self.pairs = (
            [(oo,0)] +
            [(i,oo) for i in xrange(2,self.u)] + [(i+1,i) for i in xrange(1,self.u)] +
            [(i,oo) for i in xrange(self.u+2,self.u+self.v+1)] +
            [(i+1,i) for i in xrange(self.u+1,self.u+self.v+1)]
            )
        # Note: q^b t^a really corresponds to a*s - b, in contrast to subobjects, where
        # the (-1)-shift coming from Jacobians is included.
        # This also means we don't have to manually add (-1)s for extra variables.
        self.integrand = (
            (self.u,) + (self.u-2)*(1,) + (self.u-1)*(-1,) + (2*self.v-1)*(0,),
            (self.d+1-self.v,) + (2*self.u-3)*(0,) + (self.v-1)*(-1,) + self.v*(1,)
            )

        self.datum = IgusaDatum(F + G + [LaurentIdeal(gens=[], RS=self.RS, ring=FractionField(self.ring))])
        self.datum = self.datum.simplify()

        return self.datum

    # def purge_denominator(self, denom):
    #     raise NotImplementedError
    #     return CyclotomicRationalFunction(denom.polynomial,
    #                                       [vector(ZZ,(0,-1 if denom.exponents[0][1] < 0 else 0))] + [ (a,b) for (a,b) in denom.exponents[1:] if a > 0 and b >= 0])

    # def degree_bound_in_t(self):
    #     return 0
    
    # def degree_bound_in_q(self):
    #     return 0

    def topologically_evaluate(self, shuffle=False):
        return (SR(1) + TopologicalZetaProcessor.topologically_evaluate(self, shuffle=shuffle)).factor()

    def topologically_evaluate_regular(self, datum):
        if not datum.is_regular():
            raise ValueError('need a regular datum')

        euler_cap = {}
        torus_factor_dim = {}

        N = Set(range(len(datum.polynomials)))
        
        _,d,_ = split_off_torus([datum.initials[i].num for i in N])
        min_dim = datum.ambient_dim - d

        # NOTE: triangulation/"topologisation" of RationalSet instances only
        # considers cones of maximal dimension.
        if datum.RS.dim() <= min_dim - 2:
            return
            yield

        if datum.RS.dim() >= min_dim:
            raise RuntimeError('this should be impossible')

        for I in Subsets(N):
            F = [datum.initials[i].num for i in I]
            V = SubvarietyOfTorus(F, torus_dim=datum.ambient_dim)
            U,W = V.split_off_torus()
            torus_factor_dim[I] = W.torus_dim

            assert torus_factor_dim[I] >= min_dim

            if W.torus_dim > min_dim:
                continue
            euler_cap[I] = U.khovanskii_characteristic() if U.is_nondegenerate() else U.euler_characteristic()

        for I in Subsets(N):
            chi = sum((-1)**len(J) * euler_cap[I+J] for J in Subsets(N-I) if torus_factor_dim[I+J] == min_dim)
            if not chi:
                continue
            
            I = list(I)
            id = identity_matrix(ZZ, len(I))
            def vectorise(first, k, vec):
                w = id[I.index(k)] if k in I else vector(ZZ,len(I))
                return vector(ZZ, [first] + list(vec) + list(w))
            
            polytopes = []
            for (i,j) in self.pairs:
                vertices = [vectorise(0, k, datum.ideals[i].initials[m].exponents()[0]) for m,k in enumerate(datum._ideal2poly[i])] +\
                           [vectorise(1, k, datum.ideals[j].initials[m].exponents()[0]) for m,k in enumerate(datum._ideal2poly[j])]
                polytopes.append(Polyhedron(vertices=vertices, ambient_dim=1+datum.ambient_dim+len(I)))
            extended_RS = (RationalSet(StrictlyPositiveOrthant(1)) * datum.RS *
                           RationalSet(StrictlyPositiveOrthant(len(I))))

            for surf in topologically_evaluate_monomial_integral(extended_RS, polytopes,
                                                                 self.integrand, dims=[min_dim+len(I)]):
                yield SURF(scalar=chi*surf.scalar, rays=surf.rays)

    def padically_evaluate(self, shuffle=False):
        if self.root() is None:
            return SR(1)
        return (1 + (1 - SR('q')**(-1))**(-1) * LocalZetaProcessor.padically_evaluate(self, shuffle=shuffle)).factor()

    def padically_evaluate_regular(self, datum, extra_RS=None):
        if not datum.is_regular():
            raise ValueError('need a regular datum')

        # The extra variable's valuation is in extra_RS.
        if extra_RS is None:
            extra_RS = RationalSet(StrictlyPositiveOrthant(1))
            
        q = var('q')
        count_cap = {}
        N = Set(range(len(datum.polynomials)))
        
        for I in Subsets(N):
            F = [datum.initials[i].num for i in I]
            V = SubvarietyOfTorus(F, torus_dim=datum.ambient_dim)
            count_cap[I] = V.count()

            # BEGIN SANITY CHECK
            # q = var('q')
            # u, w = V.split_off_torus()
            # assert ((count_cap[I]/(q-1)**w.torus_dim).simplify_full())(q=1) == u.euler_characteristic()
            # END SANITY CHECK

        for I in Subsets(N):
            cnt = sum((-1)**len(J) * count_cap[I+J] for J in Subsets(N-I))
            if not cnt:
                continue
            
            I = list(I)
            id = identity_matrix(ZZ, len(I))
            def vectorise(first, k, vec):
                w = id[I.index(k)] if k in I else vector(ZZ,len(I))
                return vector(ZZ, [first] + list(vec) + list(w))
            
            polytopes = []
            for (i,j) in self.pairs:
                vertices = [vectorise(0, k, datum.ideals[i].initials[m].exponents()[0]) for m,k in enumerate(datum._ideal2poly[i])] +\
                           [vectorise(1, k, datum.ideals[j].initials[m].exponents()[0]) for m,k in enumerate(datum._ideal2poly[j])]
                polytopes.append(Polyhedron(vertices=vertices, ambient_dim=1+datum.ambient_dim+len(I)))
                
            extended_RS = extra_RS * datum.RS * RationalSet(StrictlyPositiveOrthant(len(I)))

            foo, ring = symbolic_to_ratfun(cnt * (q-1)**(1+len(I))/q**(1+datum.ambient_dim), [var('t'),var('q')])
            corr_cnt = CyclotomicRationalFunction.from_laurent_polynomial(foo, ring)

            for z in padically_evaluate_monomial_integral(extended_RS, polytopes, self.integrand):
                yield corr_cnt * z

    def __repr__(self):
        if self.root() is not None:
            s = 'Representation processor:\n'
            s += 'Base ring: %s\n' % self.datum.ring
            s += 'R =\n' + str(self.R) + '\n'
            s += 'u = %d\n' % self.u
            s += 'S =\n' + str(self.S) + '\n'
            s += 'v = %d\n' % self.v
            s += 'Root:\n%s' % self.datum
            return s
        else:
            return 'Trivial representation processor'



"""
Fast arithmetic with simplicial univariate rational functions.
"""

from sage.all import *

from collections import namedtuple

from array import array

import ctypes, gzip, struct

import common

from util import create_logger, TemporaryDirectory, readable_filesize
logger = create_logger(__name__)

#
# Try to load the C cruncher. If that fails, we'll use a Python implementation.
#
from pkg_resources import resource_exists, resource_filename
try:
    if resource_exists('zetalib', 'crunch.so'):
        common.libcrunch = ctypes.CDLL(resource_filename('zetalib', 'crunch.so'))
except Exception as e:
    logger.warn('The C extension crunch.so could not be loaded. '
                'Computations of topological zeta functions will use '
                'the slower Python implementation.')
    # To debug uncomment:
    #logger.warn(e)

# [S]implicial [U]nivariate [R]ational [F]unctions are expressions of the
# form scalar / prod(rays[i][0]*s - rays[i][1], i=1,...,e).

SURF = namedtuple('SURF', ['scalar', 'rays'])
class SURFError(Exception):
    pass

_BUFSIZE = 32*1024*1024
INTSIZE = 4

if array('i').itemsize != INTSIZE:
    raise SURFError('Need sizeof(int) == 4')

class SURFSum(object):
    def __init__(self, filename, compresslevel=9):
        self._filename = filename
        self._meta_filename = filename + '.meta'
        self._compresslevel = compresslevel

        try:
            f = open(self._meta_filename, 'rb')
        except IOError:
            # No metadata -> start from scratch.
            self._critical = set()
            self._cand = dict()
            self._count = 0
            self._file = gzip.open(filename, 'wb', compresslevel)
        else:
            # Recover metadata and open file for appending.
            with f:
                D = loads(f.read())
            self._cand, self._critical, self._count = D['cand'], D['critical'], D['count']
            self._file = gzip.open(filename, 'ab', compresslevel)

    def __str__(self):
        return 'Sum of %d SURFs [%d critical points]' % (self._count, len(self._critical))

    def close(self):
        # Flush data file and write metadata.
        if not self._file.closed:
            self._file.close()
        with open(self._meta_filename, 'wb') as f:
            f.write(dumps(
                    { 'cand':     self._cand,
                      'critical': self._critical,
                      'count':    self._count }
                    ))

    def add(self, S):
        if self._file.closed:
            raise SURFError('Lost access to the SURFSum file')

        self._count += 1

        # Binary format:
        # k scalar a[0] b[0] a[1] b[1] ... a[k-1] b[k-1]

        # NOTE:
        # In rare cases, we can run into numbers that don't fit into 32-bit
        # integers. We then use an ad hoc hack to reduce to smaller numbers.
        # To trigger this, count ideals in L(6,11).

        raw = map(int, [len(S.rays), S.scalar])
        for (a,b) in S.rays:
            try:
                _ = array('i', [int(a), int(b)])
                raw.extend([int(a), int(b)])
            except OverflowError:
                if a:
                    raise SURFError('Ray does not fit into a pair of 32-bit integers')
            else:
                continue

            fac = factor(b)
            li = flatten([[p]*e for (p,e) in fac])
            li[0] *= fac.unit()
            # assert b == prod(li)
            if len(li) % 2 == 0:
                li[0] = -li[0]
            # assert -b == prod(-c for c in li)
            for c in li:
                raw.extend([int(0), int(c)])
            raw[0] += len(li)-1

        try:
            self._file.write(array('i', raw).tostring())
        except OverflowError:
            raise SURFError('Number too large to fit into a 32-bit integer')

        # Update the candidate denominator.
        E = {}
        for (a,b) in S.rays:
            if a == 0:
                continue

            self._critical.add(QQ(b)/QQ(a))

            # Only consider a*s-b with gcd(a,b) = 1 and a > 0.
            g = int(gcd((a, b)))
            a,b = a/g, b/g

            if a < 0:
                a,b = -a, -b

            # (possible, depending on the counting problem) TODO:
            # Get rid of things like s+1 (for subobjects) that cannot show up
            # in the final result.

            if E.has_key((a,b)):
                E[(a,b)] += 1
            else:
                E[(a,b)] = 1

        # Now 'E' gives the multiplicities of candidate terms for S.
        # We take the largest multiplicities over all 'S'.
        for r in E.keys():
            if (not self._cand.has_key(r)) or self._cand[r] < E[r]:
                self._cand[r] = E[r]

def _crunch_c(args):
    argc = int(len(args))
    argv = (ctypes.c_char_p*int(argc))()
    for i in xrange(argc):
        argv[i] = args[i]
    return common.libcrunch.crunch(argc,argv)

# A python implementation of `crunch.c'.
def _crunch_py(argv):

    if len(argv) < 4: return 1

    with open(argv[1], 'r') as file:
        nvalues = int(file.readline())
        values = [QQ(v) for v in file]

    if len(values) != nvalues: return 1

    results = nvalues * [QQ(0)]

    for dfile in argv[3:]:
        with gzip.open(dfile, 'rb') as gzfile:
            gzfile._read_eof = lambda *args: None # ignore gzip CRC errors in streams

            while True:
                raw = gzfile.read(INTSIZE)
                if not raw:
                    break # EOF

                if len(raw) != INTSIZE: return 1

                nrays, = struct.unpack('i', raw)
                chunk_size = 2*nrays + 1

                raw = gzfile.read(chunk_size * INTSIZE)
                if len(raw) != chunk_size * INTSIZE: return 1
                chunk = [QQ(v) for v in struct.unpack('i' * chunk_size, raw)]
                
                for i in xrange(nvalues):
                    results[i] += chunk[0] / prod(chunk[j]*values[i] - chunk[j+1] for j in xrange(1,chunk_size,2))

    with open(argv[2], 'w') as file:
        for r in results:
            file.write(str(r) + '\n')
    return 0

def crunch(args):
    return _crunch_c(args) if common.libcrunch else _crunch_py(args)

def multicrunch(surfsums, varname=None):
    """
    Given an iterable consisting of SURFSums, compute the rational function
    given by their combined sum.
    Note that this rational function necessarily has degree <= 0.
    """
    
    surfsums = list(surfsums)

    #
    # Combine the various critical sets and construct a candidate denominator.
    #

    critical = set().union(*(Q._critical for Q in surfsums))
    cand = dict()
    for Q in surfsums:
        E = Q._cand
        for r in E.keys():
            if not cand.has_key(r) or cand[r] < E[r]:
                cand[r] = E[r]

    if varname is None:
        varname = 's'

    R = QQ[varname]
    s = R.gen(0)
    g = R(prod((a*s-b)**e for ((a,b),e) in cand.iteritems()))
    m = g.degree()

    logger.info('Total number of SURFs: %d' % sum(Q._count for Q in surfsums))

    for Q in surfsums:
        Q._file.flush()

    logger.info('Combined size of data files: %s' %
         readable_filesize(sum(os.path.getsize(Q._filename) for Q in surfsums))
         )
    logger.info('Number of critical points: %d' % len(critical))
    logger.info('Degree of candidate denominator: %d' % m)

    #
    # Construct m + 1 non-critical points for evaluation.
    #

    values = set()
    while len(values) < m + 1:
        x = QQ.random_element()
        if x in critical:
            continue
        values.add(x)
    values = list(values)

    #
    # Set up parallel computations.
    #

    bucket_size = ceil(float(len(values))/common.ncpus)

    dat_filenames = [Q._filename for Q in surfsums]

    res_names = []
    val_names = []

    value_batches = [values[j::common.ncpus] for j in xrange(common.ncpus)]

    with TemporaryDirectory() as tmpdir:
        for j,v in enumerate(value_batches):
            if not v:
                break

            val_filename = os.path.join(tmpdir, 'values%d' % j)
            val_names.append(val_filename)
            res_names.append(os.path.join(tmpdir, 'results%d' % j))
            with open(val_filename, 'w') as val_file:
                val_file.write(str(len(v)) + '\n')
                for x in v:
                    val_file.write(str(x) + '\n')

        def fun(k):
            ret = crunch(['crunch', val_names[k], res_names[k]]+dat_filenames)
            if ret == 0:
                logger.info('Cruncher #%d finished.' % k)
            return ret

        fun = sage.parallel.decorate.parallel(ncpus=len(res_names))(fun)
        logger.info('Launching %d crunchers.' % len(res_names))

        for (arg, ret) in fun(range(len(res_names))):
            if ret == 'NO DATA':
                raise RuntimeError('A parallel process died')
            if ret != 0:
                raise RuntimeError('crunch failed')

        #
        # Collect results
        #
        pairs = []

        for j,rn in enumerate(res_names):
            it_batch = iter(value_batches[j])
            with open(rn, 'r') as res_file:
                for line in res_file:
                    # We also need to evaluate the candidate denominator 'g'
                    # from above at the given random points.
                    x = QQ(next(it_batch))
                    pairs.append((x, g(x)*QQ(line)))
            
    if len(values) != len(pairs):
        raise RuntimeError('Length of results is off') 

    f = R.lagrange_polynomial(list(pairs))
    res = SR(f/g)
    return res.factor() if res else res



import common
import sage.all
import sage.interfaces
import sage.parallel
import os
import sys
import itertools
import random
import struct

from sage.all import Integer, Rational, FractionField

from tmplist import DiskList
from collections import Counter

from util import create_logger
logger = create_logger(__name__)

BUCKET_SIZE = 64

import bz2
Open = bz2.BZ2File
# Open = lambda name, mode: open(name, mode, 0x100000)

from libc.stdint cimport int64_t, uint32_t, uint64_t

from sage.libs.gmp.all cimport *

cdef class HashAdder:
    cdef dict D, simple
    cdef object func

    def __init__(HashAdder self, func):
        self.D = {}
        self.simple = {}
        self.func = func

    def __len__(HashAdder self):
        return len(self.D)

    def __getitem__(HashAdder self, key):
        return self.D[key]

    def __setitem__(HashAdder self, key, value):
        self.D[key] = value
        self.simple[key] = False
        
    def __delitem__(HashAdder self, key):
        del self.D[key]
        del self.simple[key]

    def __contains__(HashAdder self, key):
        return key in self.D

    cdef add(HashAdder self, value, key=None, simple=False):
        if key is None:
            key = self.func(value)

        if key not in self.D:
            self.D[key] = value
            self.simple[key] = simple
        else:
            self.D[key] += value
            self.simple[key] = False

    cdef addmany(HashAdder self, iterable):
        for value in iterable:
            self.add(value)

    def simplify(HashAdder self):
        cdef bint done = False

        while not done:
            done = True
            for key in self.D:
                # Only keep things that evaluate to 'True'.
                if not self.D[key]:
                    del self.D[key]
                    del self.simple[key]
                    done = False
                    break  # Since we deleted a key, we should abort iterating over self.D.

                if self.simple[key]:
                    continue

                new_value = self.D[key].simplify()
                new_key = self.func(new_value)

                if new_key == key or (new_key not in self.D):
                    self.D[key] = new_value
                    self.simple[key] = True
                else:
                    del self.D[key]
                    del self.simple[key]
                    self.D[new_key] += new_value
                    self.simple[new_key] = False
                    done = False
                    break

    cdef values(HashAdder self):
        return self.D.itervalues()

    def keys(HashAdder self):
        return self.D.iterkeys()

    cdef items(HashAdder self):
        return self.D.iteritems()

    def __iter__(HashAdder self):
        return self.D.iterkeys()

# A term a * x^e[0] * y^e[1] for a = n/d (n > 0) is stored using exactly 36
# bytes as follows:
#   - e[0] + (e[1] << 16) (uint32_t)
#   - lowest, highest 64-bits of n (2 * uint64_t)
#   - lowest 64-bits of abs(d) (uint64_t)
#   - is_nonnegative(d) + (abs(highest 64-bits of abs(d))<<1) (uint64_t)
cdef enum:
        _string_length = 36

if sizeof(uintmax_t) != sizeof(uint64_t):
    raise RuntimeError("Zeta requires sizeof(uintmax_t) == sizeof(uint64_t)")

cdef class Rational256:
    cdef uint64_t n_lo, n_hi, d_lo, d_hi

    def __init__(Rational256 self, x=None):
        if x is None:
            x = Rational(0)
        n, d = long(x.numerator()), long(x.denominator())
        if n < 0:
            n = -n
            d = -d

        cdef bint is_negative = (d < 0)
        d = -d if is_negative else d

        self.n_lo = n & 0xffffffffffffffffULL # 2^64-1
        self.n_hi = n >> 64
        self.d_lo = d & 0xffffffffffffffffULL
        self.d_hi = d >> 64

        if self.d_hi & 0x8000000000000000ULL: # check if the 2^63 bit is set
            raise OverflowError('denominator does not fit into 128 bits')
        self.d_hi = (self.d_hi << 1) | (1 if is_negative else 0)

    def dumps(Rational256 self):
        return struct.pack('=QQQQ', self.n_lo, self.n_hi, self.d_lo, self.d_hi)

    def loads(Rational256 self, str s):
        self.n_lo, self.n_hi, self.d_lo, self.d_hi = struct.unpack('=QQQQ', s)

    def __nonero__(Rational256 self):
        return self.n_lo == 0 and self.n_hi == 0

    cdef to_mpq_t(Rational256 self, mpq_t x):
        cdef mpz_t a, b
        cdef bint is_negative

        cdef uint64_t d_hi_shifted = self.d_hi >> 1

        mpz_init(a) ; mpz_init(b)

        is_negative = self.d_hi & 1

        if self.n_hi:
            mpz_set_ux(a, self.n_lo)
            mpz_set_ux(b, self.n_hi)
            mpz_mul_2exp(b, b, 64)
            mpz_ior(mpq_numref(x), a, b)
        else:
            mpz_set_ux(mpq_numref(x), self.n_lo)

        if d_hi_shifted:
            mpz_set_ux(a, self.d_lo)
            mpz_set_ux(b, d_hi_shifted)
            mpz_mul_2exp(b, b, 64)
            mpz_ior(mpq_denref(x), a, b)
        else:
            mpz_set_ux(mpq_denref(x), self.d_lo)

        # mpq_canonicalize(x) # not needed!

        if is_negative:
            mpq_neg(x, x)
            
        mpz_clear(a) ; mpz_clear(b)

    cdef from_mpq_t(Rational256 self, mpq_t x):
        cdef int sgn
        cdef mpz_t a, b, c

        sgn = mpq_sgn(x)

        if not sgn:
            self.n_lo = self.n_hi = self.d_hi = 0
            self.d_lo = 1
            return

        mpz_init(a) ; mpz_init(b) ; mpz_init(c)

        #### a = abs(x.numerator())
        mpz_set(a, mpq_numref(x))
        if sgn == -1: # NOTE: denominators of canonicalised rational numbers are positive
            mpz_neg(a, a)

        #### self.n_lo = b = a & (2^64 - 1)
        mpz_set_ux(c, 0xffffffffffffffffULL)
        mpz_and(b, a, c)
        self.n_lo = mpz_get_ux(b)

        #### self.n_hi = b = a >> 64
        mpz_fdiv_q_2exp(b, a, 64)
        self.n_hi = mpz_get_ux(b)

        #### b = a >> 128
        mpz_fdiv_q_2exp(b, b, 64)
        if mpz_sgn(b):
            raise OverflowError('Numerator does not fit into 128 bits')

        #### a = x.denominator()
        mpz_set(a, mpq_denref(x))

        #### self.d_lo = b = a & (2^64 - 1)
        mpz_and(b, a, c)
        self.d_lo = mpz_get_ux(b)

        #### b = a >> 64
        mpz_fdiv_q_2exp(b, a, 64)
        if mpz_tstbit(b, 63):
            raise OverflowError('Denominator does not fit into 128 bits [sign]')

        ####  self.d_hi = b = ((a >> 64) << 1) | is_negative
        mpz_mul_2exp(b, b, 1)
        mpz_set_ui(c, 1 if sgn == -1 else 0)
        mpz_ior(b, b, c)
        self.d_hi = mpz_get_ux(b)

        ### b = a >> 128
        mpz_fdiv_q_2exp(b, b, 64)
        if mpz_sgn(b):
            raise OverflowError('Denominator does not fit into 128 bits [size]')

        mpz_clear(a) ; mpz_clear(b) ; mpz_clear(c)

    def __add__(Rational256 self, Rational256 other):
        # return Rational256(self.eval() + other.eval())
        cdef Rational256 x = Rational256.__new__(Rational256)
        cdef mpq_t a, b

        mpq_init(a) ; mpq_init(b)

        self.to_mpq_t(a)
        other.to_mpq_t(b)
        mpq_add(a, a, b)
        x.from_mpq_t(a)

        mpq_clear(a) ; mpq_clear(b)
        return x

    def eval(Rational256 self):
        cdef uint64_t d_hi = self.d_hi
        cdef bint is_negative
        cdef int64_t d64 = 0

        is_negative = d_hi & 1
        d_hi >>= 1

        if d_hi or self.d_lo & 0x8000000000000000ULL:
            d = Integer(self.d_lo) | (Integer(d_hi) << 64)
            d = -d if is_negative else d
        else:
            # +/- [d_hi|d_lo] fits into a signed(!) 64-bit integer.
            d64 = -self.d_lo if is_negative else self.d_lo
            d = Integer(d64)

        return ((Integer(self.n_lo) | (Integer(self.n_hi) << 64)) if self.n_hi else Integer(self.n_lo))/d

cdef class DictPolynomial:
    cdef dict D

    def __init__(DictPolynomial self, f=None):
        cdef uint32_t key

        self.D = {}

        if isinstance(f, dict):
            self.D = f
        elif hasattr(f, 'read'):
            self.load(f)
        elif hasattr(f, 'exponents'):
            for c, alpha in itertools.izip(f.coefficients(), f.exponents()):
                if len(alpha) != 2:
                    raise NotImplementedError('DictPolynomial only supports bivariate polynomials')
                if alpha[0] < 0 or alpha[0] > 0xffff or alpha[1] < 0 or alpha[1] > 0xffff:
                    raise ValueError('exponents do not fit into 16 bits')
                key = alpha[0] + (alpha[1] << 16)
                self.D[key] = Rational256(c)

    def _add_key_value(DictPolynomial self, uint32_t key, value):
        try:
            a = self.D[key] + value
        except KeyError:
            self.D[key] = value
            return

        if not a:
            del self.D[key]
        else:
            self.D[key] = a
                
    def __iadd__(DictPolynomial self, DictPolynomial other):
        cdef uint32_t key
        for key, value in other.D.iteritems():
            self._add_key_value(key, value)
        return self
   
    def polynomial(DictPolynomial self, ring):
        cdef uint32_t key
        y, x = ring.gens() # at some point, I decided to order variables as (t,q) instead of (q,t)
        return ring.sum(value.eval() * y**(key & 0xffff) * x**(key >> 16) for key, value in self.D.iteritems())

    def dump(DictPolynomial self, file):
        cdef uint32_t key
        for key, value in self.D.iteritems():
            file.write(struct.pack('=I', key) + value.dumps())

    def _terms_from_file(DictPolynomial self, file):
        cdef uint32_t key
        cdef Rational256 x
        cdef str raw

        while True:
            raw = file.read(_string_length)
            if not raw:
                return
            if len(raw) != _string_length:
                raise RuntimeError

            key, = struct.unpack('=I',raw[:4])
            x = Rational256()
            x.loads(raw[4:])
            yield key, x
   
    def add_from_file(DictPolynomial self, file):
        cdef uint32_t key
        cdef Rational256 value

        for key, value in self._terms_from_file(file):
            self._add_key_value(key, value)

    def load(DictPolynomial self, file):
        for key, value in self._terms_from_file(file):
            self.D[key] = value

def _addmany_numerator(filenames, denominator, *args):
    cdef str tmpdir = os.path.dirname(filenames[0])

    # Read all CRFs into memory. If that's a problem, then we're in more
    # serious trouble.
    cdef list all_crfs = list(itertools.chain.from_iterable(DiskList(f) for f in filenames))
    for f in filenames:
        DiskList(f).unlink()

    common_denominator = denominator.evaluate_reciprocal()

    def add(k):
        cdef DictPolynomial res, dp, z
        cdef int i, j, pid

        R = denominator.ring
        K = FractionField(R)
        res = DictPolynomial()

        if not common.save_memory:
            for a in itertools.islice(all_crfs, k, None, common.ncpus):
                try:
                    a_in_K =  K.coerce(a.evaluate())
                except TypeError:
                    raise ValueError('this example is not uniform: use the symbolic dispatcher instead')
                num, den = a_in_K.numerator(), a_in_K.denominator()
                res += DictPolynomial(R.product(num, common_denominator // den))
            with Open(os.path.join(tmpdir, 'res%d' % k), 'wb') as f:
                res.dump(f)
            logger.info('Summator #%d finished.' % k)
            return

        indices = range(k, len(all_crfs), common.ncpus)
        for i in xrange(0, len(indices), BUCKET_SIZE):
            pid = os.fork()
            if not pid:
                import sage.misc.misc; reload(sage.misc.misc)
                sage.interfaces.quit.invalidate_all()

                try:
                    try:
                        os.remove(os.path.join(tmpdir, 'fun%d' % k))
                    except OSError:
                        pass

                    dp = DictPolynomial()
                    for j in xrange(i, i + BUCKET_SIZE):
                        if j >= len(indices):
                            break
                        a_in_K =  K.coerce(all_crfs[indices[j]].evaluate())

                        num, den = a_in_K.numerator(), a_in_K.denominator()
                        dp += DictPolynomial(R.product(num, common_denominator // den))

                    with Open(os.path.join(tmpdir, 'fun%d' % k), 'wb') as f:
                        dp.dump(f)

                    os._exit(0)
                except Exception as e:
                    print >>sys.stderr, 'Unhandled exception in forked process:', e
                    sys.stderr.flush()
                    os._exit(1)
            else:
                _, status = os.waitpid(pid, 0)
                if status:
                    raise RuntimeError('The forked process of summator %d died with exit code %d.' % (k, status))
                with Open(os.path.join(tmpdir, 'fun%d' % k), 'rb') as f:
                    res.add_from_file(f)

        with Open(os.path.join(tmpdir, 'res%d' % k), 'wb') as f:
            res.dump(f)
        logger.info('Summator #%d finished.' % k)

    R = denominator.ring
    res = DictPolynomial()
    logger.info('Launching %d summators.' % common.ncpus)

    ET = sage.parallel.decorate.parallel(p_iter='reference' if common.debug else 'fork', ncpus=common.ncpus)(add)
    for (arg, ret) in ET(range(common.ncpus)):
        if ret == 'NO DATA':
            raise RuntimeError('A parallel process died.')
        with Open(os.path.join(tmpdir, 'res%d' % arg[0]), 'rb') as f:
            res.add_from_file(f)

    logger.info('All summators finished.')
    return res.polynomial(R) / denominator.evaluate_reciprocal()

def _addmany_symbolic(filenames, *args):
    def add(k):
        cnt = 0
        res = sage.all.SR(0)
        for f in filenames[k::common.ncpus]:
            Q = DiskList(f)
            for a in Q:
                z =  sage.all.SR(a.evaluate('symbolic'))
                z = z.factor() if z else z
                res += z
                cnt += 1
                # if cnt % 8 == 0 and res:
                #    res = res.factor()
        res = res.factor() if res else res
        logger.info('Summator #%d finished.' % k)
        return res

    res = sage.all.SR(0)
    logger.info('Launching %d summators.' % common.ncpus)
    ET = sage.parallel.decorate.parallel(p_iter='reference' if common.debug else 'fork', ncpus=common.ncpus)(add)
    for (arg, ret) in ET(range(common.ncpus)):
        if ret == 'NO DATA':
            raise RuntimeError('A parallel process died.')
        res += ret
    logger.info('All summators finished.')
    return res

def addmany(filenames, denominator):
    dispatcher = { 'symbolic': _addmany_symbolic, 'numerator': _addmany_numerator }
    method = common.addmany_dispatcher
    if method not in dispatcher:
         raise ValueError('invalid dispatcher')
    logger.info('addmany dispatcher: %s' % method)
    return dispatcher[method](filenames, denominator)  # (filenames, denominator if purged_denominator is None else purged_denominator, bound_t, bound_q)

# Merge CRFs with the same denominator.    
def _optimise_match(filenames):
    logger.info('Total number of rational functions: %d' % sum(len(DiskList(f)) for f in filenames))

    tmpdir = os.path.dirname(filenames[0])

    Q = []
    fp = lambda cr: tuple(tuple(e) for e in cr.exponents[1:])
    stats = Counter(fp(cr) for cr in itertools.chain.from_iterable(DiskList(f) for f in filenames))

    logger.debug('Total number of denominators = %d, maximal multiplicity = %d, number of singletons = %d' % (len(stats),
                                                                                                             max(stats.values()),
                                                                                                             sum(1 for v in stats if stats[v] == 1)))
    logger.debug('Joining rational functions.')
    D = {}

    new_filenames = [os.path.join(tmpdir, 'opt_match_eval%d' % k) for k in xrange(common.ncpus)]
    Q = [DiskList(f) for f in new_filenames]

    idx = 0
    for cr in itertools.chain.from_iterable(DiskList(f) for f in filenames):
        thumb = fp(cr)
        if thumb not in D:
            D[thumb] = cr
        else:
            D[thumb] += cr

        # NOTE: we don't take into account that after summation and
        # cancellation, the fingerprint might change. We simply add
        # all CRFs with the same fingerprint together (detecting zero).

        if not D[thumb]:
            del D[thumb]

        stats[thumb] -= 1

        if not stats[thumb]:
            del stats[thumb]
            if thumb in D:
                Q[idx].append(D[thumb])
                del D[thumb]
            idx = (idx + 1) % common.ncpus

    for f in filenames:
        DiskList(f).unlink()

    logger.debug('New number of rational functions: %d' % sum(1 for cr in itertools.chain.from_iterable(DiskList(f) for f in new_filenames)))
    return new_filenames

# Add CRFs according to various criteria.
def _optimise_combine(filenames):
    logger.info('Total number of rational functions: %d' % sum(len(DiskList(f)) for f in filenames))
    logger.debug('Reading all rational functions into system memory.')
    fingerprint = lambda cr: tuple(tuple(e) for e in cr.exponents[1:])

    HA = HashAdder(fingerprint)
    HA.addmany(itertools.chain.from_iterable(DiskList(f) for f in filenames))

    HA.simplify()

    def judge(r):
        return sum(1 for x in HA if r in x)

    def process(HA, sorter, selector):
        cnt = 0
        while True:
            cnt += 1

            for ray in sorter(HA):
                keys = [x for x in HA if ray in x]
                if len(keys) > 1 and selector(HA, ray, keys):
                    break
            else:
                break

            # If necessary, reduce the number of CRFs to be added in one
            # step by taking a random choice of at most 42 (magic constant!)
            # elements.
            indices = range(len(keys))
            random.shuffle(indices)
            indices = indices[:42]
            keys = [keys[i] for i in indices]

            logger.debug('Chosen ray %s, attained %d times.' % (ray, len(keys)))
            key = keys[0]
            value = HA[key]
            for x in keys[1:]:
                HA[key] += HA[x]
                del HA[x]
            
            if cnt % 3 == 0:
                cnt = 0
                HA.simplify()
            logger.debug('New total number: %d.' % len(HA))

    HA.simplify()
    rays_of = lambda HA: list(set(x for v in HA.keys() for x in v))
    
    def multiplicity_sorter_increasing(HA):
        rays = rays_of(HA)
        rays.sort(key=lambda ray: sum(1 for x in HA if ray in x))
        return rays

    def weight_sorter_decreasing(HA):
        rays = rays_of(HA)
        rays.sort(key=lambda x: -abs(x[0])-abs(x[1]))
        return rays

    def sum_sorter_increasing(HA):
        rays = rays_of(HA)
        rays.sort(key=lambda x: x[0] + x[1])
        return rays

    # def multiplicity_sorter_decreasing(HA):
    #     rays = multiplicity_sorter_increasing(HA)
    #     rays.reverse()
    #     return rays

    # def weight_sorter_increasing(HA):
    #     rays = rays_of(HA)
    #     rays.sort(key=lambda x: abs(x[0]) + abs(x[1]))
    #     return rays

    def pred_sum_sorter_increasing(HA, ray, keys):
        return ((ray[0] < 0) or (ray[1] < 0)) and len(keys) <= 1024 and len(keys) <= 64 * common.optimisation_level

    def pred_not_too_many(HA, ray, keys):
        return len(keys) <= 64 * common.optimisation_level

    strategies =  [
        ( 'Combining rays with negative entries and moderate multiplicities',
           sum_sorter_increasing, pred_sum_sorter_increasing ),

        ( 'Combining rays with low multiplicities',
          multiplicity_sorter_increasing, pred_not_too_many ),
        
        ( 'Combining rays with high weights and moderate multiplicities',
          weight_sorter_decreasing, pred_not_too_many ),
        ]

    for (descr, sorter, selector) in strategies:
        logger.debug('Optimising: %s.' % descr)
        process(HA, sorter, selector)
       
    logger.debug('Writing back CRFs.')
    tmpdir = os.path.dirname(filenames[0])
    new_filenames = [os.path.join(tmpdir, 'opt_combine_eval%d' % k) for k in xrange(common.ncpus)]
    Q = [DiskList(f) for f in new_filenames]

    idx = 0
    for cr in HA.values():
        if not cr:
            continue
        Q[idx].append(cr)
        idx = (idx + 1) % common.ncpus

    for f in filenames:
        DiskList(f).unlink()

    return new_filenames

def optimise(filenames):
    for opt in [_optimise_match, _optimise_combine]:
        filenames = opt(filenames)
    return filenames




"""Utility functions that don't belong anywhere else.
"""

from sage.all import *
from itertools import chain
from distutils.spawn import find_executable
from pkg_resources import resource_exists, resource_filename

import os

import logging
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)

def create_logger(name, level=logging.WARNING):
    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.propagate = False
    if not logger.handlers:
        handler.setFormatter(logging.Formatter('%(asctime)s %(name)s %(levelname)s - %(message)s', '%Y-%m-%d %H:%M:%S'))
        logger.addHandler(handler)
    return logger

logger = create_logger(__name__)

# Decorator for functions taking a single argument 'self'.
# After the first call, the cached result is stored as an
# attribute of 'self'.

def cached_simple_method(f, name=None):
    if name is None:
        name = '_cached_' + f.__name__

    def cached_f(self):
        try:
            return getattr(self, name)
        except AttributeError:
            res = f(self)
            setattr(self, name, res)
            return res
    return cached_f

def monomial_log(g):
    if not (g.is_constant() or g.is_monomial()):
        raise ValueError('Can only take logs of monomials')
    if g.denominator() == Integer(1):
        return vector(g.numerator().exponents()[0])
    else:
        return vector(g.numerator().exponents()[0]) - vector(g.denominator().exponents()[0])

def monomial_exp(R, v):
    return prod(map(lambda w: w[0]**w[1], zip(R.gens(),list(v))))

def split_vector(v):
    vpos = vector( x if x > Integer(0) else Integer(0) for x in v)
    vneg = vector(-x if x < Integer(0) else Integer(0) for x in v)
    return (vpos, vneg)

def E(n, i, j, ring=QQ):
    A = Matrix(ring, n)
    A[i,j] = 1
    return A

def unzip(li):
    return tuple(zip(*li))

def normalise_poly(f):
    return f/f.lc()

def initial_form_by_direction(f, y):
    """Determine the initial form of 'f' in direction 'y'.
    """

    mon, coeff = f.monomials(), f.coefficients()
    weights = [monomial_log(m)*y for m in mon]
    w = min(weights)
    return sum(coeff[k]*mon[k] for k in xrange(len(mon)) if weights[k] == w)

def vertex_by_direction(polytope, y):
    """Return one(!) vertex on the face of 'polytope' defined by
    minimising 'y'.
    """
    if polytope.is_empty():
        raise ValueError('need a non-empty polytope')
    vertices = polytope.vertices_matrix().transpose()
    w = min(x*y for x in vertices)
    return next(x for x in vertices if x*y == w)
    
def is_subpolynomial(f, a):
    """Test if `a' is a sum of terms of the polynomial `f'.
    """
    a = f.parent()(a)
    return len(f.monomials()) == len((f-a).monomials()) + len(a.monomials())

def subpolynomial_meet(f,g):
    """Find the 'largest' common subpolynomial of 'f' and 'g'.
    """
    mon_f, coeff_f = f.monomials(), f.coefficients()
    mon_g, coeff_g = g.monomials(), g.coefficients()

    r = f.parent()(0)
    for i in xrange(len(mon_f)):
        if mon_f[i] in mon_g:
            k = mon_g.index(mon_f[i])
            if coeff_f[i] == coeff_g[k]:
                r += coeff_f[i] * mon_f[i]
    return r

def normalise_laurent_polynomial(f):
    """
    Rescale a Laurent polynomial by Laurent monomials. Details TBA.
    """

    # Since Sage's Laurent polynomials are useless, we just use rational
    # functions instead.
    # First make sure, 'f' really is one.

    R = f.parent()
    K = FractionField(R)
    f = K(f)
    if K.ngens() == 0:
        return R(0) if not f else R(1)

    f = f.numerator()
    return R(f / gcd(f.monomials()))

def terms_of_polynomial(f):
    return [c*t for c,t in zip(f.coefficients(), f.monomials())]

def evaluate_polynomial(f, v):
    if not f:
        return f.parent().zero()
    exp = lambda e: prod(x ** e for (x,e) in zip(v,e))
    return sum(scalar * exp(e) for (scalar, e) in zip(f.coefficients(), f.exponents()))

import contextlib, shutil, tempfile

@contextlib.contextmanager
def TemporaryDirectory(delete=True):
    name = tempfile.mkdtemp()
    try:
        yield name
    finally:
        if delete:
            shutil.rmtree(name)

@contextlib.contextmanager
def cd(dir):
    oldcwd = os.getcwd()
    try:
        os.chdir(dir)
        yield
    finally:
        os.chdir(oldcwd)

def readable_filesize(size):
    return '%.1f %s' % ((size/1024.0, 'KB') if size <= 512*1024
                        else ( (size/1024.0/1024.0, 'MB'))
                        if size <= 512*1024*1024 else (size/1024.0/1024.0/1024.0, 'GB'))
              
def my_find_executable(name):
    s = os.path.join('bin', name)
    try:
        if resource_exists('zetalib', s):
            return os.path.abspath(resource_filename('zetalib', s))
    except:
        pass

    filename = find_executable(name)
    return os.path.abspath(filename) if filename else None

def minimal_elements(P, lte):
    # Let (P,lte) be a preordered set. Find a set of representatives
    # of the minimal elements in the quotient poset.

    P = list(P)
    active = range(len(P))
    for i in xrange(len(P)):
        if i not in active:
            continue
        for j in active[:]:
            if i == j:
                continue
            if lte(P[i], P[j]):
                active.remove(j)
    return [P[i] for i in active]

def squarefree_part(f):
    # NOTE: f.factor().unit() is discarded
    return f.parent().prod(h for h,_ in f.factor()) if f else f

def split_off_torus(F):
    # Given a list F of polynomials, return G,d,T where len(F) == len(G)
    # and G consists of polynomials in d variables; T = change of basis.
    # There exists A in GL(n,ZZ) s.t. F[i]^A == G[i] * (Laurent term).
    # In particular, F is non-degenerate relative to {0} iff this the case
    # for G; cf Remark 4.3(ii) and Lemma 6.1(ii) in arXiv:1405.5711.
    #
    # The number of variables in G is the dimension of Newton(prod(F)).
    #

    if not F:
        return [], 0, None

    R = F[0].parent()
    v = {f:vector(ZZ,f.exponents()[0]) for f in F}
    submodule = (ZZ**R.ngens()).submodule(chain.from_iterable((vector(ZZ,e)-v[f] for e in f.exponents()) for f in F))
    _,_,T = matrix(ZZ, submodule.basis_matrix()).smith_form()
    d = submodule.rank()
    Rd = PolynomialRing(QQ, d, R.gens()[:d])
    K = FractionField(R)
    return [Rd(normalise_laurent_polynomial(K(f([monomial_exp(K,e) for e in T.rows()])))) for f in F], d, T

def principal_minors(A, size):
    m = A.nrows()
    n = A.ncols()
    r = min(m, n)
    li = []
    for idx in Subsets(range(r), size):
        idx = list(idx)
        li.append(A.base_ring()(A.matrix_from_rows_and_columns(idx, idx).determinant()))
    return li

def upper_triangular_matrix(R, d, entries):
    entries = iter(entries)
    M = matrix(R, d, d)
    for i in xrange(d):
        for j in xrange(i, d):
            M[i,j] = next(entries)
    return M

def MyCompositions(n,length):
    if n < length or not length:
        return
    if length==1:
        yield [n]
        return

    for i in xrange(1, n-length+2):
        for c in MyCompositions(n-i, length-1):
            yield [i] + c
          
def common_overring(R, S):
    if R.has_coerce_map_from(S):
        return R
    elif S.has_coerce_map_from(R):
        return S
    else:
        combined_vars = list(set(R.base_ring().gens()) | set(S.base_ring().gens()))
        K = FractionField(PolynomialRing(QQ, len(combined_vars), combined_vars))
        return PolynomialRing(K, R.gens())

def is_block_diagonal_matrix(A, blocks):
    if not A.is_square():
        raise ValueError

    partial_sums = [sum(blocks[:i]) for i in xrange(len(blocks)+1)]
    n = A.nrows()
    if partial_sums[-1] != n:
        raise ValueError

    for r in xrange(len(partial_sums)-1):
        a = partial_sums[r]
        b = partial_sums[r+1]

        for i in xrange(a,b):
            for j in xrange(b,n):
                if A[i,j] != 0 or A[j,i] != 0:
                    return False
    return True

def symbolic_to_polynomial(f, vars):
    # Rewrite a symbolic expression as a polynomial over SR in a given
    # list of variables.

    poly = f.polynomial(QQ)
    allvars = poly.variables()

    indices = []
    for x in allvars:
        try:
            indices.append(vars.index(x))
        except ValueError:
            indices.append(None)

    R = PolynomialRing(SR, len(vars), vars)
    res = R.zero()
    for coeff, alpha in zip(poly.coefficients(), poly.exponents()):
        if type(alpha) == int:
            # Once again, univariate polynomials receive special treatment
            # in Sage.
            alpha = [alpha]

        term = R(coeff)
        for i, e in enumerate(alpha):
            if not e:
                continue
            if indices[i] is None:
                term *= SR(allvars[i]**e)
            else:
                term *= R.gen(indices[i])**e
        res += term
    return res

def symbolic_to_ratfun(f, vars):
    R = PolynomialRing(QQ, len(vars), vars)
    try:
        return R(f.numerator())/R(f.denominator()), R
    except TypeError:
        pass

    num = symbolic_to_polynomial(f.numerator(), vars)
    den = symbolic_to_polynomial(f.denominator(), vars)
    return num/den, num.parent()

def augmented_env(filename):
    env = os.environ.copy()
    env['PATH'] = os.path.dirname(filename) + os.path.pathsep + env.get('PATH','')
    return env

def evaluate_matrix(A, y, ring=None):
    if ring is None:
        ring = y[0].parent()
    return matrix(ring, [[entry(list(y)) for entry in rows] for rows in A])
    
def basis_of_matrix_algebra(gens, product=None):
    # Given a list of rational square matrices, compute
    # a basis of the rational matrix algebra they generate.
    
    if (product is None) or (product == 'zero'):
        mult = lambda a,b: 0*a
    elif product == 'associative':
        mult = lambda a,b: a*b
    elif product == 'Lie':
        mult = lambda a,b: a*b - b*a
    else:
        mult = product

    if not gens:
        return []

    n = gens[0].nrows()
    V = QQ**(n**2)
    U = V.subspace([g.list() for g in gens])

    while True:
        W = U
        for b in U.basis():
            for c in U.basis():
                v = vector(mult(Matrix(QQ, n, n, list(b)),
                                Matrix(QQ, n, n, list(c))).list())
                if v not in U:
                    W = W + V.subspace([v])
        if W == U:
            break
        else:
            U = W

    def normalise(w):
        c = lcm([a.denominator() for a in w])
        return [ZZ(c*a) for a in w]
    return [ Matrix(QQ, n, n, normalise(list(b))) for b in U.basis() ]

def matlist_to_mat(mats, n=None):
    if not mats and n is None:
        raise ValueError('cannot guess size of matrices')
    R = PolynomialRing(QQ, 'y', len(mats))
    if not mats:
        return matrix(R, n, n)
    return matrix(R, sum(R.gen(i) * a for i,a in enumerate(mats)))

def mat_to_matlist(A):
    return [matrix(QQ, [[a(list(e)) for a in rows] for rows in A]) for e in identity_matrix(QQ, A.base_ring().ngens())]



from sage.all import *

import abstract
from abstract import ZetaDatum, TopologicalZetaProcessor, LocalZetaProcessor, ReductionError

from toric import ToricDatum

from torus import SubvarietyOfTorus, CountException
from convex import conify_polyhedron, DirectProductOfPolyhedra, StrictlyPositiveOrthant, RationalSet
from triangulate import topologise_cone
from surf import SURF
from cycrat import CyclotomicRationalFunction
from util import normalise_poly, terms_of_polynomial, symbolic_to_ratfun

import itertools

from util import create_logger
from reps import padically_evaluate_monomial_integral

logger = create_logger(__name__)

BALANCE_FULL_BOUND = 10
DEPTH_BOUND = 24

class Breakout(Exception):
    pass

def null_chopper(T, indices):
    raise ReductionError('The null chopper never works')

def equality_chopper(T, indices, depth_bound=None):
    # Only decompose the polyhedron, let simplify() take care of the polynomials.
    # We discard empty stuff so calling this function multiple times does no
    # harm.

    if depth_bound is None:
        depth_bound = DEPTH_BOUND

    if T._depth > depth_bound:
        raise ReductionError('Bound for the depth exceeded. Equality chopper failed.')

    lhs = T.lhs
    rhs = T.initials if T.is_balanced() else T.rhs

    for i,j in itertools.combinations(indices,2):
        f = normalise_poly(rhs[i] * lhs[j])
        g = normalise_poly(rhs[j] * lhs[i])
        q = f/g
        if (f == g) or (normalise_poly(q.numerator()).is_monomial() and
                        normalise_poly(q.denominator()).is_monomial()):
            logger.debug('Reduction using the pair (%d,%d)' % (i,j))
            lt, rt = [T.reduce(i,j).simplify(), T.reduce(j,i,strict=True).simplify()]
            lt._depth += 1 ; rt._depth += 1
            logger.info('Left depth increased to %d' % lt._depth)
            logger.info('Right depth increased to %d' % rt._depth)
            return [lt, rt]
    raise ReductionError('Equality chopper failed')

def greedy_chopper(T, indices, depth_bound=None, use_first_improvement=False):
    if depth_bound is None:
        depth_bound = DEPTH_BOUND

    if not T.is_balanced():
        raise ReductionError('Can only handle balanced toric data')

    if len(T._coll_idx) == 1:
         raise ReductionError('Toric singularities!')

    # To ensure termination, we bound the depth of reduction steps
    # that don't lead to immediate improvements.

    inweight = lambda T: sum(len(r.monomials())-1 for r in T.initials)

    def cand(i,j,ti,tj):
        # Assigns a `badness' value between 0.0 and Infinity to a possible
        # reduction step.

        Ti = T.reduce(i,j,ti,tj).simplify()
        Tj = T.reduce(j,i,tj,ti,strict=True).simplify()

        children = [t.simplify() for t in itertools.chain(Ti.balance(), Tj.balance())]
        naughty = []
        for t in children:
            if t.is_regular():
                continue
            if inweight(t) >= inweight(T):
                t._depth += 1
                logger.info('Depth increased to %d' % t._depth)
                naughty.append(t)

        if not naughty:
            return float(0), children

        if any(t._depth > depth_bound for t in naughty):
            logger.info('Exceeded the given bound for the depth of reductions.')
            return Infinity, children

        return (float(sum(inweight(t) for t in naughty))/ (inweight(T) * len(naughty)),
                naughty + [c for c in children if c not in naughty])

    # NOTE: the partition of the polyhedron is the same for any choice of (i,j)
    # so it would suffice to compute it once for every pair (i,j).

    optval, optsol, reduction_quad = Infinity, None, None
    try:
        for i,j in itertools.combinations(indices, 2):
            for ti,tj in itertools.product(terms_of_polynomial(T.initials[i]),
                                           terms_of_polynomial(T.initials[j])):
                val, sol = cand(i,j,ti,tj)
                if val < optval:
                    optval, optsol = val, sol
                    reduction_quad = (i,j,ti,tj)
                if not optval or (use_first_improvement and optval < 1.0):
                    raise Breakout()
    except Breakout:
        pass
    if optsol is None:
        raise ReductionError('Greedy chopper failed')
    logger.info('Greedy chopper: badness after reduction %.2f' % optval)
    logger.debug('Greedy chopper: reduction candidate used is (%d, %d, %s, %s)' % (i,j,ti,tj))
    return optsol

class ReductionStrategy(object):
    def __init__(self, name, chopper, prechopper, order):
        self.name = name
        self.chopper = chopper
        self.prechopper = prechopper
        self.order = order

    def __str__(self):
        return 'ReductionStrategy:%s' % self.name

class Strategy(object):
    NORMAL = ReductionStrategy('normal', greedy_chopper, null_chopper, False)
    ORDER = ReductionStrategy('order', greedy_chopper, null_chopper, True)
    PREEMPTIVE = ReductionStrategy('preemptive', greedy_chopper, equality_chopper, True)
    NONE = ReductionStrategy('none', null_chopper, null_chopper, False)

class SubobjectDatum(ZetaDatum):
    # Just a thin wrapper on top of ToricDatum from Zeta 0.1.
    def __init__(self, T, strategy=None):
        self.toric_datum = T
        self.strategy = Strategy.NORMAL if strategy is None else strategy

    def is_empty(self):
        return self.toric_datum.is_empty()

    def simplify(self):
        return SubobjectDatum(self.toric_datum.simplify(), strategy=self.strategy)

    def is_balanced(self):
        return self.toric_datum.is_balanced()
    
    def balance(self):
        balancing_strategy = 'full' if self.toric_datum.weight() < BALANCE_FULL_BOUND else 'min'
        for B in self.toric_datum.balance(strategy=balancing_strategy):
            yield SubobjectDatum(B, strategy=self.strategy)

    def is_ordered(self):
        return (not self.strategy.order) or self.toric_datum.is_ordered()

    def order(self):
        for O in self.toric_datum.order():
            yield SubobjectDatum(O, strategy=self.strategy)

    def is_regular(self):
        return self.toric_datum.is_regular()

    def __repr__(self):
        return str(self.toric_datum)

    def reduce(self, preemptive=False):
        chops = self.strategy.prechopper(self.toric_datum, range(len(self.toric_datum.rhs))) if preemptive else self.strategy.chopper(self.toric_datum, self.toric_datum._coll_idx)
        for C in chops:
            yield SubobjectDatum(C, strategy=self.strategy)
            
class SubobjectZetaProcessor(TopologicalZetaProcessor, LocalZetaProcessor):
    def __init__(self, algebra, objects, strategy=None):
        self.algebra = algebra
        self.objects = objects
        self.strategy = strategy
        self._root = None

    def root(self):
        if self._root is None:
            self._root = SubobjectDatum(self.algebra.toric_datum(self.objects), strategy=self.strategy)
        return self._root

    def optimise(self):
        A = self.algebra.find_good_basis(self.objects)
        self.algebra = self.algebra.change_basis(A)
        self._root = None

    def topologically_evaluate_regular(self, datum):
        T = datum.toric_datum
        if not T.is_regular():
            raise ValueError('Can only processed regular toric data')

        # All our polyhedra all really half-open cones (with a - 1 >=0
        # being an imitation of a >= 0).

        C = conify_polyhedron(T.polyhedron)

        M = Set(range(T.length()))

        logger.debug('Dimension of polyhedron: %d' % T.polyhedron.dim())

        # STEP 1:
        # Compute the Euler characteristcs of the subvarieties of
        # Torus^sth defined by some subsets of T.initials.
        # Afterwards, we'll combine those into Denef-style Euler characteristics
        # via inclusion-exclusion.

        logger.debug('STEP 1')
        s = SR.var('s')

        alpha = {} 
        tdim = {}

        for I in Subsets(M):
            logger.debug('Processing I = %s' % I)
            F = [T.initials[i] for i in I]

            V = SubvarietyOfTorus(F, torus_dim=T.ambient_dim)
            U,W = V.split_off_torus()

            # Keep track of the dimension of the torus factor for F == 0.
            tdim[I] = W.torus_dim

            if tdim[I] > C.dim():
                # In this case, we will never need alpha[I].
                logger.debug('Totally irrelevant intersection.')
                # alpha[I] = ZZ(0)
            else:
                # To ensure that the computation of Euler characteristics succeeds in case
                # of global non-degeneracy, we test this first.
                # The 'euler_characteristic' method may change generating sets,
                # possibly introducing degeneracies.
                alpha[I] = U.khovanskii_characteristic() if U.is_nondegenerate() else U.euler_characteristic()
                logger.debug('Essential Euler characteristic alpha[%s] = %d; dimension of torus factor = %d' % (I,alpha[I], tdim[I]))

        logger.debug('Done computing essential Euler characteristics of intersections: %s' % alpha)

        # STEP 2:
        # Compute the topological zeta functions of the extended cones.
        # That is, add extra variables, add variable constraints (here: just >= 0),
        # and add newly monomialised conditions.

        Z = {}

        def cat(u,v):
            return vector(list(u) + list(v))

        logger.debug('STEP 2')
        for I in Subsets(M):
            logger.debug('Current set: I = %s' % I)

            # P = C_0 x R_(>0)^I in the paper
            P = DirectProductOfPolyhedra(T.polyhedron, StrictlyPositiveOrthant(len(I)))

            it = iter(identity_matrix(ZZ, len(I)).rows())
            ieqs = []
            for i in M:
                # Turn lhs[i] | monomial of initials[i] * y[i] if in I,
                #      lhs[i] | monomial of initials[i] otherwise
                # into honest cone conditions.
                ieqs.append(
                    cat(vector(ZZ,(0,)),
                        cat(
                            vector(ZZ, T.initials[i].exponents()[0]) -
                            vector(ZZ, T.lhs[i].exponents()[0]),
                            next(it) if i in I else zero_vector(ZZ,len(I))
                            )
                        )
                    )

            if not ieqs:
                # For some reason, not providing any constraints yields the empty
                # polyhedron in Sage; it should be all of RR^whatever, IMO.
                ieqs = [ vector(ZZ, (T.ambient_dim+len(I)+1) * [0]) ]

            Q = Polyhedron(ieqs=ieqs, base_ring=QQ, ambient_dim=T.ambient_dim+len(I))

            sigma = conify_polyhedron(P.intersection(Q))
            logger.debug('Dimension of Hensel cone: %d' % sigma.dim())

            # Obtain the desired Euler characteristic via inclusion-exclusion,
            # restricted to those terms contributing to the constant term mod q-1.
            chi = sum((-1)**len(J)*alpha[I+J] for J in Subsets(M-I) if tdim[I+J] + len(I) == sigma.dim())

            if not chi:
                continue

            # NOTE: dim(P) = dim(sigma): choose any point omega in P
            # then a large point lambda in Pos^I will give (omega,lambda) in sigma.
            # Moreover, small perturbations of (omega,lambda) don't change that
            # so (omega,lambda) is an interior point of sigma inside P.

            surfs = (topologise_cone(sigma, matrix([ 
                            cat(T.integrand[0], zero_vector(ZZ,len(I))),
                            cat(T.integrand[1], vector(ZZ, len(I)*[-1]))
                            ]).transpose())
                     )

            for S in surfs:
                yield SURF(scalar=chi*S.scalar, rays=S.rays)

    # def purge_denominator(self, denom):
    #     raise NotImplementedError
    #     return CyclotomicRationalFunction(denom.polynomial,
    #                                       [vector(ZZ,(0,-self.algebra.rank))] + [ (a,b) for (a,b) in denom.exponents[1:] if a > 0 and b >= 0])

    def padically_evaluate(self, shuffle=False):
        q = var('q')
        res = q**self.algebra.rank / (q - 1)**self.algebra.rank * LocalZetaProcessor.padically_evaluate(self, shuffle=shuffle)
        return res.factor() if res else res

    def padically_evaluate_regular(self, datum):
        T = datum.toric_datum
        if not T.is_regular():
            raise ValueError('Can only processed regular toric data')

        M = Set(range(T.length()))
        q = SR.var('q')

        alpha = {} 

        for I in Subsets(M):
            F = [T.initials[i] for i in I]
            V = SubvarietyOfTorus(F, torus_dim=T.ambient_dim)
            alpha[I] = V.count()

        Z = {}

        def cat(u,v):
            return vector(list(u) + list(v))

        for I in Subsets(M):
            cnt = sum((-1)**len(J)*alpha[I+J] for J in Subsets(M-I))
            if not cnt:
                continue

            P = DirectProductOfPolyhedra(T.polyhedron, StrictlyPositiveOrthant(len(I)))

            it = iter(identity_matrix(ZZ, len(I)).rows())
            ieqs = []
            for i in M:
                ieqs.append(
                    cat(vector(ZZ,(0,)),
                        cat(
                            vector(ZZ, T.initials[i].exponents()[0]) -
                            vector(ZZ, T.lhs[i].exponents()[0]),
                            next(it) if i in I else zero_vector(ZZ,len(I))
                            )
                        )
                    )

            if not ieqs:
                ieqs = [ vector(ZZ, (T.ambient_dim+len(I)+1) * [0]) ]

            Q = Polyhedron(ieqs=ieqs, base_ring=QQ, ambient_dim=T.ambient_dim+len(I))

            foo, ring = symbolic_to_ratfun(cnt * (q-1)**len(I)/q**(T.ambient_dim), [var('t'),var('q')])
            corr_cnt = CyclotomicRationalFunction.from_laurent_polynomial(foo, ring)

            Phi = matrix([ cat(T.integrand[0], zero_vector(ZZ,len(I))),
                           cat(T.integrand[1], vector(ZZ, len(I)*[-1])) ]).transpose()
            sm = RationalSet([P.intersection(Q)]).generating_function()
            for z in sm.monomial_substitution(QQ['t','q'], Phi):
                yield corr_cnt * z
        
    def __repr__(self):
        return 'Subobject zeta processor\nAlgebra:\n%s\nObjects: %s\nRoot:\n%s' % (self.algebra, self.objects, self.root())




"""
Closed subvarieties of algebraic tori.
"""

from sage.all import *

from itertools import chain
from convex import mixed_volume
from util import normalise_laurent_polynomial, squarefree_part, monomial_exp, \
    terms_of_polynomial, cached_simple_method, split_off_torus

from common import symbolic_count_varieties

import common

from util import create_logger, MyCompositions

logger = create_logger(__name__)

def symbolic_variable(V):
    try:
        idx = symbolic_count_varieties.index(V)
    except ValueError:
        idx = len(symbolic_count_varieties)
        symbolic_count_varieties.append(V)
    return var('sc_%d' % idx)

# def monomial_sat(I):
#     R = I.ring()
#     J = R.ideal(prod(R.gens()))
#     sat, _ = singular.sat(I._singular_(), J._singular_())
#     _, sat = singular.mstd(sat)
#     return R.ideal(sat)

class CountException(Exception):
    pass

class SubvarietyOfTorus(object):
    def __init__(self, polynomials=None, torus_dim=None):
        if not polynomials:
            if torus_dim is None:
                raise ValueError('Need to specify the dimension')
            self.torus_dim = torus_dim
            self.ring = PolynomialRing(QQ, self.torus_dim, 'x')
            self.polynomials = []
            return

        R = polynomials[0].parent()
        if (torus_dim is not None) and (torus_dim != R.ngens()):
            raise ValueError('Supplied dimension does not match the given polynomials')
        self.torus_dim = R.ngens()
        self.ring = PolynomialRing(QQ, self.torus_dim, 'x')

        if self.torus_dim == 0:
            # Some things like GCDs fail for polynomial rings in 0 variables. *sigh*
            self.polynomials = [self.ring.one()] if any(f for f in polynomials) else []
            return

        # Map everything into our 'reference polynomial ring' and normalise.
        theta = R.hom(self.ring.gens())
        nm = lambda f: normalise_laurent_polynomial(f/f.lc())
        polynomials = [nm(theta(f)) for f in polynomials if f != 0]
        self.polynomials = [self.ring.one()] if (1 in polynomials) else polynomials
        return

    def __eq__(self, other):
        if not self.torus_dim == other.torus_dim:
            return False
        return sorted(self.polynomials) == sorted(other.polynomials)

    def _solvable_conditions(self):
        """
        Produce all (i,x,g) such that F[i] == 0 is equivalent to x == g;
        here x is one of the defining variables and g is a Laurent polynomial
        which only involves variables != x.
        """

        vars = self.ring.gens()
        F = self.polynomials

        for i in xrange(len(F)):
            for x in vars:
                terms = terms_of_polynomial(F[i])
                cand = [t for t in terms if x.divides(t)]
                if len(cand) != 1:
                    continue

                t0 = cand[0]
                if (x**2).divides(t0):
                    continue

                g = -sum(t for t in terms if t != t0)/(t0//x)
                logger.debug('[%d] %s == 0 is equivalent to %s == %s' % (i, F[i], x, g))
                yield (i, x, g)

    def _isolated_variables(self):
        """
        Produce all (i,x,u,v) such that
            (1) F[i] == 0 is equivalent to u == v * x,
            (2) x does not occur in u, v, nor in F[j] for i != j.
        """
        
        F = self.polynomials
        for x in self.ring.gens():
            seen = False
            cand = None

            for i,f in enumerate(F):
                vterms = [t//x for t in terms_of_polynomial(f) if x.divides(t)]
                if not vterms:
                    continue

                if seen or any(x.divides(t) for t in vterms):
                    cand = None
                    break

                seen = True
                v = sum(vterms)
                cand = (i,x,v*x-f,v)
            if not cand is None:
                yield cand

    @cached_simple_method
    def split_off_torus(self):
        """
        Given non-zero polynomials F defining a a subvariety V of T^n,
        we seek to find polynomials G in d variables such that
        V ~ W * T^(n-d), where W is defined by G. Returns W and T^(n-d).
        The number d is the dimension of the Minkowski sum of the Newton
        polytopes of the polynomials in F.
        """
        G,d, _ = split_off_torus(self.polynomials)
        return SubvarietyOfTorus(G, torus_dim=d), SubvarietyOfTorus(torus_dim=self.torus_dim-d)

    def simplify_defining_equations(self):
        F = self.polynomials[:]

        if self.torus_dim == 1:
            F = [gcd(F)]
        F = [squarefree_part(f) for f in F]

        # # If the saturated generating set is 'smaller', use that.
        # G = monomial_sat(self.ring.ideal(F)).gens()
        # if sum(len(f.coefficients()) for f in G) < sum(len(f.coefficients()) for f in F):
        #   logger.debug('Using saturated generators.')
        #    F = G

        # 'mu' measures the 'size' of a polynomial for our greedy algorithm.
        mu = lambda f: len(f.coefficients()) 
        changed = True
        while changed:
            changed = False

            for i in xrange(len(F)):
                if F[i] == 0:
                    continue
                if F[i] == 1 or F[i].is_monomial():
                    logger.debug('Variety is empty.')
                    return SubvarietyOfTorus(polynomials=[self.ring.one()],
                                             torus_dim=self.torus_dim)
                for j in xrange(len(F)):
                    if i == j:
                        continue

                    if F[j] == 0:
                        continue

                    # Important invariant: non-zero polynomials in F are monic.

                    # First attempt: Polynomial division.
                    _, r = F[i].quo_rem(F[j])
                    r = squarefree_part(r)

                    if mu(r) < mu(F[i]):
                        logger.debug('(quo/rem) Replacing F[%d]=%s by %s' % (i,F[i],r))
                        F[i] = r if not r else normalise_laurent_polynomial(r/r.lc())
                        changed = True
                        continue

                    # Second attempt: reduction.
                    # We try ALL conceivable pairs.
                    reduction_performed = True
                    while reduction_performed:
                        reduction_performed = False
                        for ti in terms_of_polynomial(F[i]):
                            for tj in terms_of_polynomial(F[j]):
                                g = gcd(ti,tj)

                                # BEGIN HACK
                                #if g == 1:
                                #    continue
                                # END HACK

                                r = (tj//g) * F[i] - (ti//g) * F[j]
                                r = squarefree_part(r)

                                if mu(r) < mu(F[i]):
                                    logger.debug('(reduce) Replacing F[%d]=%s by %s' % (i,F[i],r))
                                    logger.debug('(reduce) Using term %s of F[%d] and term %s of F[%d]'
                                         % (ti,i,tj,j))
                                    F[i] = r if not r else normalise_laurent_polynomial(r/r.lc())
                                    changed = True
                                    reduction_performed = True
                                    break
                            else:
                                continue
                            break

        F = [f for f in F if f != 0]
        return SubvarietyOfTorus(polynomials=F, torus_dim=self.torus_dim)

    @cached_simple_method
    def khovanskii_characteristic(self):
        """
        Given k polytopes P[0], ..., P[k-1] in n-space, compute the sum 
        (-1)^(n+k) * n! * MV(P[0], ..., P[0], P[1], ..., P[1], ...)
        over all compositions of n.
        This number is an integer which is equal to the Euler characteristic
        of the subvariety of TT^n defined by sufficiently non-degenerate
        (Laurent) polynomials with Newton polytopes P[0], ..., P[k-1].
        """

        if not self.polynomials:
            return 1 if self.torus_dim == 0 else 0
        elif any(f.is_constant() and f != 0 for f in self.polynomials):
            return 0

        P = [f.newton_polytope() for f in self.polynomials]
        k = len(P)

        n = P[0].ambient_dim()
        if any(q.ambient_dim() != n for q in P):
            raise TypeError('All polytopes must have the same ambient dimension')

        if k > n:
            return 0
        
        return (
            (-1)**(n+k) * factorial(n) *
            sum(mixed_volume(chain.from_iterable([q] * a for (q, a) in zip(P, c)))
                for c in MyCompositions(n, length=k)
                )
            )

    @cached_simple_method
    def is_nondegenerate(self):
        if not self.polynomials:
            return True
        from toric import is_nondegenerate
        return is_nondegenerate(self.polynomials, all_subsets=False, all_initial_forms=True)

    def _count_general(self, level):
        # Levels:
        # -1 - Euler characteristic
        #  0 - polynomial in q
        #  1 - polynomial in q or roots of univariate polynomials
        #  2 - general closed subvarieties of tori

        euler = level < 0 # == only compute euler characteristic 
        q = int(1) if euler else var('q')

        logger.debug('Initial polynomials: %s' % self.polynomials)
        
        V,W = self.simplify_defining_equations().split_off_torus()
       
        if W.torus_dim > 0:
            logger.debug('Split off torus factor of dimension %d' % W.torus_dim)

        logger.debug('Simplified polynomials: %s' % self.polynomials)

        if W.torus_dim > 0:
            return 0 if euler else V._count_general(level) * (q-1)**W.torus_dim

        V = V.simplify_defining_equations()
        if not V.polynomials:
            return (q-1)**V.torus_dim # note: 0**0 == 1
        elif any(f.is_constant() and f != 0 for f in V.polynomials):
            return 0 if euler else SR(0) # empty set
        elif euler and V.is_nondegenerate():
            logger.debug('The variety is Khovanskii-non-degenerate')
            return V.khovanskii_characteristic()

        if not euler and V.torus_dim == 1:
            assert len(V.polynomials) == 1
            f = V.polynomials[0]
            if len(f.factor()) == f.degree(): # check if f splits completely over QQ
                return SR(f.degree())
            if level == 0:
                raise CountException('cannot handle number fields when level == 0')
            else:
                return symbolic_variable(V)

        logger.debug('Trying to decompose the variety...')
        F = V.polynomials
        I = range(len(F))

        for (i,x,g) in V._solvable_conditions():
            # NOTE: We need to specify the number of variables in the following
            # line for otherwise Sage might use a UNIVARIATE polynomial ring;
            # these behave differently.
            S = PolynomialRing(QQ, V.torus_dim-1, [y for y in V.ring.gens() if y != x])
            li = [S(normalise_laurent_polynomial(F[j].subs({x: g})))
                  for j in I if j != i]
            U = SubvarietyOfTorus(li, V.torus_dim-1)
            W = SubvarietyOfTorus([S(normalise_laurent_polynomial(g))] + li)
            try:
                res = U._count_general(level) - W._count_general(level)
            except CountException:
                pass
            else:
                logger.debug('Successfully decomposed variety.')
                logger.debug('Solvable condition: (%d, %s, %s)' % (i,x,g))
                logger.debug('U = %s' % U.polynomials)
                logger.debug('W = %s' % W.polynomials)
                return res

        logger.debug('Trying to isolate a variable...')
        for (i,x,u,v) in V._isolated_variables():
            # When F[i] == 0 is equivalent to x*v==u and x doesn't occur in
            # any other F[j], there are two cases:
            # (1) On the subvariety v==0, x is arbitrary so this part has Euler
            # characteristic zero.
            # (2) On v != 0, x == u/v is completely determined by the
            # remaining variables and the implicit condition u != 0.

            S = PolynomialRing(QQ, V.torus_dim-1, [y for y in V.ring.gens() if y != x])
            li = [S(F[j]) for j in I if j != i]

            U = SubvarietyOfTorus(li, V.torus_dim-1)
            Wu = SubvarietyOfTorus([S(u)]+li, V.torus_dim-1)
            Wv = SubvarietyOfTorus([S(v)]+li, V.torus_dim-1)
            Z = SubvarietyOfTorus([S(u), S(v)] + li, V.torus_dim-1)

            try:
                res = U._count_general(level) - Wu._count_general(level) - Wv._count_general(level) + q * Z._count_general(level)
            except CountException:
                pass
            else:
                logger.debug('Succesfully isolated a variable.')
                logger.debug('Isolated variable data: (%d, %s, %s, %s)' % (i,x,u,v))
                logger.debug('U = %s' % U.polynomials)
                logger.debug('Wu = %s' % Wu.polynomials)
                logger.debug('Wv = %s' % Wv.polynomials)
                logger.debug('Z = %s' % Z.polynomials)
                return res

        logger.debug('SubvarietyOfTorus._count_general failed. Defining polynomials: %s' % self.polynomials)

        if level < 2:
            raise CountException('failed to %s' % ('compute Euler characteristic' if euler else 'count points'))
        else:
            return symbolic_variable(self)

    @cached_simple_method
    def euler_characteristic(self):
        return self._count_general(level=-1)
            
    @cached_simple_method
    def count(self):
        for i in ([0,1,2] if common.symbolic else [0]):
            try:
                return SR(self._count_general(level=i)).expand()
            except CountException:
                logger.debug('count (level=%d) failed: %s' % (i, self))
        raise CountException('Failed to count number of rational points. Try switching to symbolic mode.')
        
    def __str__(self):
        if self.polynomials:
            return 'Subvariety of %d-dimensional torus defined by %s' % (self.torus_dim, self.polynomials)
        else:
            return 'Torus of dimension %d' % self.torus_dim

    __repr__ = __str__
 
def Torus(n):
    return SubvarietyOfTorus(torus_dim=n)



#
# Zeta.
#
# Copyright 2014, 2015, 2016, 2017 Tobias Rossmann.
#
# This package is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but without any warranty; without even the implied warranty of
# merchantability or fitness for a particular purpose.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this software; if not, see <http://www.gnu.org/licenses>.
#

from functools import partial
import pkg_resources

import common

import addmany

import examples
from subobjects import SubobjectZetaProcessor
from reps import RepresentationProcessor

from ask import AskProcessor

from algebra import Algebra
from abstract import ReductionError
from toric import ToricDatum
from util import create_logger
from common import symbolic_count_varieties

import surf
import smurf
import abstract
import subobjects
import reps
import ask
import util
import cycrat
import triangulate

from laurent import LaurentPolynomial

from subobjects import Strategy

from sage.all import Infinity, var
import sage

__version__ = pkg_resources.get_distribution("zetalib").version
__date__ = 'July 2018'

logger = create_logger(__name__)

from reps import IgusaProcessor

# The following is supposed to detect a bug which is present (at least)
# in version 7.5.1 and 7.6 of Sage.
__SERIES_BUG = False
def __check_sage_bugs():
    x,y,z = var('x y z')
    if y.series(x) != (z.series(x))(z=y):
        __SERIES_BUG = True

__check_sage_bugs()

class Profile(object):
    SAVE_MEMORY = 1
    NORMAL = 2
    SPEED = 3

def lookup(entry=None, what=None, type='topological'):
    if entry is None:
        return [(i, examples.names[i][:]) for i, _ in enumerate(examples.algebras)]

    i = examples.id.get(entry)
    if i is None:
        try:
            i = int(entry)
        except ValueError:
            raise KeyError('Unknown algebra')
    if i < 0 or i >= len(examples.algebras):
        raise KeyError('Invalid index')

    if what is None:
        return examples.algebras[i]
    elif what == 'id':
        return i
    elif what == 'names':
        return examples.names[i]

    if type == 'topological':
        D = examples.topzetas
    elif type in ['p-adic', 'local']:
        D = examples.padzetas
    else:
        raise ValueError('unknown type')
    return D[i].get(what, None)

def zeta_function(type, L, objects=None, optimise_basis=False,
                  ncpus=None, alt_ncpus=None, strategy=None, profile=None, verbose=False,
                  optlevel=None, addmany_dispatcher=None, mode=None, debug=None):
    if type not in ['p-adic', 'topological']:
        raise ValueError('Unknown type of zeta function')

    if type == 'p-adic':
        if common.count is None:
            raise RuntimeError('LattE/count is required in order to compute p-adic zeta functions')
        elif __SERIES_BUG:
            raise RuntimeError('power series expansions in this version of Sage cannot be trusted')

    # Multiprocessing.
    if ncpus is None:
        ncpus = Infinity
    from multiprocessing import cpu_count
    common.ncpus = min(ncpus, cpu_count())

    if alt_ncpus is None:
        alt_ncpus = common.ncpus
    common._alt_ncpus = alt_ncpus

    if addmany_dispatcher is None:
        addmany_dispatcher = 'numerator'
    common.addmany_dispatcher = addmany_dispatcher

    common.debug = False if debug is None else debug

    if optlevel is None:
        optlevel = 1
    common.optimisation_level = optlevel

    # Reduction strategies.
    if strategy is None:
        strategy = Strategy.NORMAL

    # Memory profiles.
    if profile is None:
        profile = Profile.NORMAL
    if profile not in [Profile.SAVE_MEMORY, Profile.NORMAL, Profile.SPEED]:
        raise ValueError('Invalid profile')
    if profile == Profile.SAVE_MEMORY:
        common.save_memory = True
        common.plumber = True
    elif profile == Profile.NORMAL:
        common.save_memory = False
        common.plumber = True
    elif profile == Profile.SPEED:
        common.save_memory = False
        common.plumber = False

    if verbose:
        from logging import INFO, DEBUG
        loglevels = [(logger, INFO), (smurf.logger, INFO), (surf.logger, INFO),
                     (torus.logger, INFO), (abstract.logger, INFO),
                     (cycrat.logger, INFO), (triangulate.logger, INFO),
                     (reps.logger, INFO), (subobjects.logger, INFO),
                     (ask.logger, INFO),
                     (addmany.logger, INFO), ]
        oldlevels = []

        for m,level in loglevels:
            old = m.getEffectiveLevel()
            oldlevels.append(old)
            m.setLevel(min(old,level))

    # Try to detect polynomials and lists of polynomials.
    polynomials = None
    try:
        _ = L.exponents()
    except:
        try:
            if not isinstance(L, basestring) and not isinstance(L,sage.matrix.matrix_dense.Matrix_dense) and objects != 'orbits':
                polynomials = list(L)
        except:
            pass
    else:
        polynomials = [L]

    if (polynomials is None) and not isinstance(L, algebra.Algebra) and not isinstance(L,sage.matrix.matrix_dense.Matrix_dense) and objects != 'orbits':
        L = lookup(L)

    if polynomials is not None:
        proc = IgusaProcessor(*polynomials)
    elif objects in ['subalgebras', 'ideals']:
        proc = SubobjectZetaProcessor(L, objects, strategy=strategy)
    elif objects == 'reps':
        proc = RepresentationProcessor(L)
    elif objects == 'ask':
        proc = AskProcessor(L, mode=mode)
    elif objects == 'orbits':
        # NOTE: we don't currently check if L really spans a matrix Lie algebra
        proc = AskProcessor(util.matlist_to_mat(util.basis_of_matrix_algebra(L, product='Lie')), mode=mode)
    elif objects == 'cc':
        if not L.is_Lie() and not L.is_nilpotent():
            raise ValueError('need a nilpotent Lie algebra in order to enumerate conjugacy classes')
        proc = AskProcessor(util.matlist_to_mat(L._adjoint_representation()), mode=mode)
    else:
        raise ValueError('unknown objects [%s]' % objects)

    if optimise_basis:
        logger.info('Searching for a good basis...')
        proc.optimise()
        logger.info('Picked a basis.')

    if verbose:
        print proc

    try:
        if type == 'p-adic':
            return proc.padically_evaluate(shuffle=True)
        elif type == 'topological':
            return proc.topologically_evaluate(shuffle=True)
    finally:
        if verbose:
            for ((m,_),level) in zip(loglevels,oldlevels):
                m.setLevel(level)

def do(type, L, objects='subalgebras', filename=None, save_memory=None, symbolic=False, **kwargs):
    if save_memory is None:
        save_memory = bool(type == 'p-adic')

    D = { 'optimise_basis': True,
          'strategy': Strategy.NORMAL if type == 'p-adic' else Strategy.PREEMPTIVE,
          'profile': Profile.SAVE_MEMORY if save_memory else Profile.NORMAL,
          'verbose': True,
          }

    if symbolic:
        common.symbolic = True
        D.update({ 'debug': True,
                   'addmany_dispatcher': 'symbolic',
                   'optlevel': 0.2 })
    else:
        common.symbolic = False

    D.update(kwargs)
    Z = zeta_function(type, L, objects, **D)
    if filename is True:
        if not isinstance(L, basestring):
            raise ValueError
        filename = objects + '-' + L + '.' + {'p-adic': 'pad', 'topological': 'top'}[type]
    if filename:
        with open(filename, 'w') as f: f.write(str(Z))
    return Z

topological_zeta_function = partial(do, 'topological', strategy=Strategy.NORMAL, verbose=False)
local_zeta_function = partial(do, 'p-adic', strategy=Strategy.NORMAL, verbose=False)

top = partial(do, 'topological')
pad = partial(do, 'p-adic')

def check(name, objects='subalgebras', type='p-adic', **kwargs):
    L = lookup(name)
    W = lookup(name, objects, type)
    if W is None:
        raise RuntimeError('zeta function not contained in database')

    Z = do(type, L, objects, **kwargs)
   
    if Z != W:
        print 'Computed: ', Z
        print 'Database: ', W
        raise RuntimeError('computed zeta function of %s differs from entry in the database' % name)

    print
    print 'Algebra #%d' % lookup(name,'id')
    print 'Names: %s' % ', '.join(lookup(name,'names'))
    print 'Objects:', objects
    print 'Type:', type
    print 'Zeta:', Z
    print 'Confirmed.'
    return True

def checkall(objects='reps', type='p-adic', **kwargs):
    ids = []

    for i, _ in lookup():
        if lookup(i, objects, type) is not None:
            check(i, objects, type, **kwargs)
            ids.append(i)

    print
    print 'Objects:', objects
    print 'Type:', type
    print 'Confirmed zeta functions for the following algebras:'

    for i in ids:
        print '#%d:\t%s' % (i, ', '.join(lookup(i, 'names')))
    print 'Total number:', len(ids)
    
if __SERIES_BUG:
    logger.critical(
            """This version of Sage may compute incorrect symbolic power series
expansions. Computations of p-adic zeta functions are therefore
unavailable. Try using a different version of Sage such as 7.4.""")

    
_banner_text = """
ZZZZZZZZZZZZZZZZZZZ                           tttt                           
Z:::::::::::::::::Z                        ttt:::t                           
Z:::::::::::::::::Z                        t:::::t                           
Z:::ZZZZZZZZ:::::Z                         t:::::t                           
ZZZZZ     Z:::::Z     eeeeeeeeeeee   ttttttt:::::ttttttt     aaaaaaaaaaaaa   
        Z:::::Z     ee::::::::::::ee t:::::::::::::::::t     a::::::::::::a  
       Z:::::Z     e::::::eeeee:::::et:::::::::::::::::t     aaaaaaaaa:::::a 
      Z:::::Z     e::::::e     e:::::tttttt:::::::tttttt              a::::a 
     Z:::::Z      e:::::::eeeee::::::e     t:::::t             aaaaaaa:::::a 
    Z:::::Z       e:::::::::::::::::e      t:::::t           aa::::::::::::a 
   Z:::::Z        e::::::eeeeeeeeeee       t:::::t          a::::aaaa::::::a 
ZZZ:::::Z     ZZZZe:::::::e                t:::::t    ttttta::::a    a:::::a 
Z::::::ZZZZZZZZ:::e::::::::e               t::::::tttt:::::a::::a    a:::::a 
Z:::::::::::::::::Ze::::::::eeeeeeee       tt::::::::::::::a:::::aaaa::::::a 
Z:::::::::::::::::Z ee:::::::::::::e         tt:::::::::::tta::::::::::aa:::a
ZZZZZZZZZZZZZZZZZZZ   eeeeeeeeeeeeee           ttttttttttt   aaaaaaaaaa  aaaa

%s
%s
                             by   
                                 Tobias Rossmann
""" % ('{:>77}'.format('VERSION ' + __version__),
       '{:>77}'.format('Released: ' + __date__))

def banner():
    print(_banner_text)



"""
Triangulations of pointed cones; includes yet another interface to Normaliz.
"""

from sage.all import *
from sage.features import Executable

import common
from collections import namedtuple
from shutil import rmtree

import subprocess, os

from util import TemporaryDirectory, cd, my_find_executable, augmented_env

from tmplist import TemporaryList

from util import create_logger
logger = create_logger(__name__)

# TODO: Consider using the Sage Normaliz interface that uses PyNormaliz.
_normaliz_exe = Executable('Normaliz',
                            executable='normaliz',
                            spkg='normaliz',
                            url='https://www.normaliz.uni-osnabrueck.de')
if not _normaliz_exe.is_present():
    logger.warn('Normaliz not found. '
                'Triangulation will be slow. '
                'You can try to install it by running "sage -i normaliz"')

# A simplicial cone 'sc' can be turned into an honest one
# via Cone(rays=sc.rays, lattice=ZZ**ambdim, check=False)).

SCONE = namedtuple('SCONE', ['rays', 'multiplicity'])
from surf import SURF

def _explode_vector(v):
    if len(v) == 0:
        return ''
    elif len(v) == 1:
        return str(v[0])
    else:
        return ' '.join('%d' % (a,) for a in v[:-1]) + ' ' + str(v[-1])

def _normalize_cone(C):
    # Produce valid normaliz input from cone.
    s = '%d\n%d\n' % (C.nrays(), C.lattice_dim())
    for r in C.rays():
        s += _explode_vector(r) + '\n'
    return s + 'integral_closure\n'

def _triangulate_cone_internal(C, all_cones=False):
    rays = list(C.rays())
    if any(a < 0 for r in rays for a in r):
        raise ValueError('Internal triangulation only implemented for non-negative cones.')
    normalised_rays = [Integer(1)/v.norm(1) * vector(QQ,v) for v in rays]

    with TemporaryList() as maximal_faces:
        for indices in PointConfiguration(normalised_rays).triangulate() if len(normalised_rays) > 1 else [[0]]:
            subrays = [rays[i] for i in indices]
            if all_cones:
                maximal_faces.append(indices)
            else:
                multiplicity = prod(a for a in matrix(ZZ,subrays).elementary_divisors() if a != 0)
                yield SCONE(rays=subrays, multiplicity=multiplicity)

        if all_cones:
            X = SimplicialComplex(maximal_faces=maximal_faces, maximality_check=False)
            for f in X.face_iterator():
                yield SCONE(rays = [rays[i] for i in f], multiplicity=-1)

def _triangulate_cone_normaliz(C, all_cones):
    if all_cones:
        raise NotImplementedError

    with TemporaryDirectory() as tmpdir:
        logger.debug('Temporary directory of this normaliz session: %s' % tmpdir)

        play = os.path.join(tmpdir, 'play')
        with open(play + '.in', 'w') as f:
            f.write(_normalize_cone(C))

        with cd(tmpdir), open('/dev/null', 'w') as DEVNULL:
            retcode = subprocess.call(
                [common.normaliz, '-B', '-t', '-T', '-x=1', play],
                stdout=DEVNULL, stderr=DEVNULL,
                env=augmented_env(common.normaliz)
                )
        if retcode != 0:
            raise RuntimeError('Normaliz failed')

        with TemporaryList() as maximal_faces:
            # Recover all the rays from the triangulation
            with TemporaryList() as rays:
                with open(play+'.tgn', 'r') as f:
                    nrays = int(f.readline())
                    ambdim = int(f.readline())
                    rays = []
                    if ambdim != C.lattice_dim():
                        raise RuntimeError('The triangulation lives in the wrong space')
                    for line in f:
                        rays.append([int(w) for w in line.split()])

                    if len(rays) != nrays:
                        raise RuntimeError('Number of rays is off')
                    logger.debug('Total number of rays in triangulation: %d' % nrays)

                # Recover the simplicial pieces
                cnt = 0
                with open(play+'.tri', 'r') as f:
                    ncones = int(f.readline())
                    mplus1 = int(f.readline())

                    if mplus1 != C.dim() + 1:
                        raise RuntimeError('.tri and .tgn files do not match')
                    for line in f:
                        if line == 'plain\n':
                            continue
                        elif line == 'nested\n':
                            raise RuntimeError('Normaliz returned a nested triangulation; this should never happen')

                        cnt += 1
                        li = [int(w) for w in line.split()]
                        if li[-1] == -1:
                            raise RuntimeError('Multiplicity is missing from Normaliz output')

                        if all_cones:
                            maximal_faces.append([i-1 for i in li[:-1]])
                            # Note that we discard all multiplicities in this case.
                        else:
                            yield SCONE(rays = [rays[i-1] for i in li[:-1]],
                                        multiplicity = li[-1]
                                        )

            logger.debug('Total number of cones in triangulation: %d' % cnt)
            if cnt != ncones:
                raise RuntimeError('Number of simplicial cones is off')

            if all_cones:
                X = SimplicialComplex(maximal_faces=maximal_faces, maximality_check=False)
                for f in X.face_iterator():
                    yield SCONE(rays = [rays[i] for i in f], multiplicity=-1)

def triangulate_cone(C, all_cones=False):
    if all_cones:
        raise NotImplementedError
    return _triangulate_cone_normaliz(C, all_cones) if common.normaliz else _triangulate_cone_internal(C, all_cones)

def _topologise_scone(sc, Phi):
    if not sc.rays:
        raise ValueError('Cannot handle empty scones')
    if (Phi.nrows() != len(sc.rays[0])) or (Phi.ncols() != 2):
        raise ValueError('Invalid substitution matrix')

    li = [(int(Phi.column(0)*vector(ZZ,v)),
           int(Phi.column(1)*vector(ZZ,v))) for v in sc.rays]

    # An element of 'li' of the form (0,b) is really just the
    # scalar -b in disguise (mind the sign!).
    # Collect all of those together, cancelling common factors
    # with the numerator sc.multiplicity.

    rays = []
    num, den = sc.multiplicity, 1
    for (a,b) in li: # this becomes 1/(a*s-b) in the SURF
        if a == 0:
            den *= (-b)
        else:
            rays.append((a,b))
    g = int(gcd(sc.multiplicity, den))
    num /= g
    den /= g

    if den == -1:
        num, den = -num, 1

    if den != 1:
        rays.append((0,-den))
    return SURF(scalar=num, rays=rays)

def topologise_cone(C, Phi):
    """
    Turn the cone 'C' into a generator of SURFs via the substitution matrix
    'Phi'.
    """

    if C.dim() == 0:
        raise RuntimeError
        # return iter([])
    else:
        return (_topologise_scone(sc, Phi) for sc in triangulate_cone(C))



from sage.all import *

from abc import ABCMeta, abstractmethod
from tmplist import TemporaryList, DiskList

from util import TemporaryDirectory, readable_filesize

import addmany

import surf
import common
import random

import itertools
from cycrat import CyclotomicRationalFunction

from contextlib import closing
from util import create_logger
logger = create_logger(__name__)

class ReductionError(Exception):
    pass

class ZetaDatum(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def is_empty(self):
        pass

    @abstractmethod
    def simplify(self):
        pass

    @abstractmethod
    def is_balanced(self):
        pass

    @abstractmethod
    def balance(self):
        return
        yield

    def is_ordered(self):
       return True

    def order(self):
       raise NotImplementedError

    @abstractmethod
    def is_regular(self):
        pass

    @abstractmethod
    def reduce(self, preemptive=False):
        pass

class GeneralZetaProcessor(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def root(self):
        pass

    def process(self, regular):
        with TemporaryList() as unprocessed:
            r = self.root()
            if r is not None:
                unprocessed.append(r)

            good = 0
            while unprocessed:
                datum = unprocessed.pop()
                logger.info('POP [#unprocessed = %2d, #regular = %d]' % (len(unprocessed)+1,len(regular)))

                datum = datum.simplify()

                if datum.is_empty():
                    continue

                if datum.is_regular():
                    good += 1
                    regular.append(datum)
                    continue

                # NOTE:
                # One could argue that there's really only ONE operation which
                # might act as balancing, ordering, reduction etc.
                #

                def do(things, name):
                    cnt = 0
                    for B in things:
                        if B.is_empty():
                            continue
                        unprocessed.append(B)
                        cnt += 1
                    logger.info('%s performed. Pushed %d new cases.' % (name, cnt))
                    
                try:
                    chops = list(datum.reduce(preemptive=True))
                except ReductionError:
                    pass
                else:
                    do(chops, 'Preemptive reduction')
                    continue

                if not datum.is_balanced():
                    logger.info('Balancing...')
                    do(datum.balance(), 'Balancing')
                    continue

                if not datum.is_ordered():
                    logger.info('Ordering...')
                    do(datum.order(), 'Ordering')
                    continue

                try:
                    chops = datum.reduce(preemptive=False)
                except ReductionError, e:
                    logger.debug("Don't know how to reduce this.")
                    raise e
                else:
                    do(chops, 'Reduction')
                    continue

        return regular

    def optimise(self):
        pass

class TopologicalZetaProcessor(GeneralZetaProcessor):
    __metaclass__ = ABCMeta

    @abstractmethod
    def topologically_evaluate_regular(self, datum):
        pass

    def topologically_evaluate(self, shuffle=False):
        with TemporaryList() as regular:
            #
            # Main loop.
            #
            logger.info('Beginning main loop.')
            self.process(regular)
            logger.info('Main loop finished successfully.')
            logger.info('Total number of regular data: %d' % len(regular))

            #
            # Evaluate.
            #
            with TemporaryDirectory() as evaldir:
                surfsum_names = [os.path.join(evaldir, 'surfsum%d' % k) for k in xrange(common.ncpus)]

                indices = range(len(regular))
                if shuffle:
                    random.shuffle(indices)

                def fun(k):
                    with closing(surf.SURFSum(surfsum_names[k])) as Q:
                        for i in xrange(k, len(regular), common.ncpus):
                            for S in self.topologically_evaluate_regular(regular[indices[i]]):
                                Q.add(S)
                    logger.info('Evaluator #%d finished.' % k)

                ET = sage.parallel.decorate.parallel(p_iter='reference' if common.debug else 'fork', ncpus=common.ncpus)(fun)
                logger.info('Launching %d evaluators.' % common.ncpus)
                for (arg, ret) in ET(range(common.ncpus)):
                    if ret == 'NO DATA':
                        raise RuntimeError('A parallel process died.')
                logger.info('All evaluators finished.')

                #
                # Crunch.
                #
                surfsums = []
                try:
                    surfsums = [surf.SURFSum(name) for name in surfsum_names]
                    return surf.multicrunch(surfsums)
                finally:
                    for Q in surfsums:
                        Q.close()

class LocalZetaProcessor(GeneralZetaProcessor):
    __metaclass__ = ABCMeta

    @abstractmethod
    def padically_evaluate_regular(self, datum):
        pass

    #def purge_denominator(self, denom):
    #    return denom

    # def degree_bound_in_t(self):
    #     """Returns an upper bound in 't' for the degree of a generic local zeta function.
    #     """
    #     return +Infinity

    # def degree_bound_in_q(self):
    #     """Returns an upper bound in 'q' for the degree of a generic uniform local zeta function.
    #     """
    #     return +Infinity

    def padically_evaluate(self, shuffle=False):
        with TemporaryList() as regular:
            logger.info('Beginning main loop.')
            self.process(regular)
            logger.info('Main loop finished successfully.')
            logger.info('Total number of regular data: %d' % len(regular))

            with TemporaryDirectory() as tmpdir:
                eval_filenames = [os.path.join(tmpdir, 'eval%d' % k) for k in xrange(common.ncpus)]
                def evaluate(k):
                    def fun(i):
                        try:
                            Q = DiskList(eval_filenames[k])
                            for a in self.padically_evaluate_regular(regular[i]):
                                Q.append(a)
                        except Exception as e: # For debugging, change this to 'except RuntimeError as E'.
                            return e
                        else:
                            return True

                    if common.plumber and not common.debug:
                        fun = sage.parallel.decorate.fork(fun)
                            
                    for i in xrange(k, len(regular), common.ncpus):
                        e = fun(i)
                        if e is not True:
                            raise e # This will result in 'NO DATA' being returned below.
                    logger.info('Evaluator #%d finished.' % k)

                ET = sage.parallel.decorate.parallel(p_iter='reference' if common.debug else 'fork', ncpus=common.ncpus)(evaluate)
                logger.info('Launching %d evaluators.' % common.ncpus)
                for (arg, ret) in ET(range(common.ncpus)):
                    if ret == 'NO DATA':
                        raise RuntimeError('A parallel process died.')
                logger.info('All evaluators finished.')

                eval_filenames = addmany.optimise(eval_filenames)
                logger.info('Total number of rational functions: %d' % sum(len(DiskList(f)) for f in eval_filenames))

                if common.addmany_dispatcher != 'symbolic':
                    denom = CyclotomicRationalFunction.common_denominator(itertools.chain.from_iterable(DiskList(f) for f in eval_filenames))
                    logger.info('Degree of denominator in (t,q): (%d,%d)'  % (-denom.degree(0), -denom.degree(1)))
                else:
                    denom = None

                # if common.addmany_dispatcher.find('interpol') != -1:
                #     purged_denom = self.purge_denominator(denom)
                #     logger.info('Degree of purged denominator in (t,q): (%d,%d)' % (-purged_denom.degree(0), -purged_denom.degree(1)))

                #     bound_t, bound_q = -Infinity, -Infinity
                #     for cr in itertools.chain.from_iterable(DiskList(f) for f in eval_filenames):
                #         bound_t = max(bound_t, cr.degree(0))
                #         bound_q = max(bound_q, cr.degree(1))
                #     logger.debug('bound_t = %d, bound_q = %d' % (bound_t,bound_q))
                # else:
                #     purged_denom = None
                #     bound_t = None
                #     bound_q = None

                ncpus = common.ncpus
                if common._alt_ncpus is not None:
                    common.ncpus = common._alt_ncpus

                try:
                    res = addmany.addmany(eval_filenames, denom)
                finally:
                    common.ncpus = ncpus

                with open(os.path.join(tmpdir, 'unfactored'), 'w') as f:
                    f.write(str(res)) # just in case factoring runs out of memory

                # Final recovery.
                logger.info('Factoring final result.')

                if res.parent() == SR:
                    res = res.factor() if res else res
                elif res:
                    fac = res.factor()
                    res = SR(fac.unit()) * SR.prod(SR(a)**i for a,i in fac)
                logger.info('All done.')
                return res



from sage.all import *

import common

from util import create_logger, cached_simple_method

logger = create_logger(__name__)

from reps import IgusaDatum, RepresentationProcessor
from convex import PositiveOrthant, StrictlyPositiveOrthant, RationalSet
from util import evaluate_matrix
from laurent import LaurentIdeal, LaurentPolynomial
from torus import SubvarietyOfTorus
from abstract import LocalZetaProcessor

class AskProcessor(RepresentationProcessor):
    def __init__(self, arg, mode=None):
        if mode is None:
            mode = 'O'

        if mode not in ['O', 'K']:
            raise ValueError("supported modes are 'O' and 'K'")
        
        self.R = arg
        self.d = self.R.nrows()
        self.e = self.R.ncols()
        self.mode = mode

        self.ring = self.R.base_ring()
        self.ell = self.ring.ngens()

        self.basis = [evaluate_matrix(self.R, y) for y in identity_matrix(QQ, self.ell)]
        V = (QQ**(self.d * self.e)).subspace([A.list() for A in self.basis])
        if V.dimension() != self.ell:
            raise ValueError('subspace parameterised by matrix of linear forms has wrong dimension')

    def padically_evaluate_regular(self, datum):
        # The essential change from representation zeta functions is that
        # the extra variable may be a unit here.
        for z in super(AskProcessor, self).padically_evaluate_regular(datum, extra_RS=RationalSet(PositiveOrthant(1))):
            yield z

    def topologically_evaluate_regular(self, datum):
        raise NotImplementedError

    @cached_simple_method
    def root(self):
        if self.mode == 'K':
            ell = self.ell
            self.RS = RationalSet([PositiveOrthant(ell)], ambient_dim=ell)

            actual_ring = self.ring
            F = FractionField(actual_ring)
            r = matrix(F, self.R).rank()
            self.r = r
            if not self.d:
                return

            F = [
                LaurentIdeal(
                    gens = [LaurentPolynomial(f) for f in self.R.minors(j)],
                    RS = self.RS,
                    normalise = True)
                for j in range(0, r+1)
            ]
        elif self.mode == 'O':
            self.RS = RationalSet([PositiveOrthant(self.d)], ambient_dim=self.d)
            actual_ring = PolynomialRing(QQ, 'x', self.d)
            xx = vector(actual_ring, actual_ring.gens())
            self.C = matrix(actual_ring, [xx * matrix(actual_ring,A) for A in self.basis])
            r = matrix(FractionField(actual_ring), self.C).rank()
            self.r = r
            if not self.d:
                return
            F = [
                LaurentIdeal(
                    gens = [LaurentPolynomial(f) for f in self.C.minors(j)],
                    RS = self.RS,
                    normalise = True)
                for j in range(0, r+1)
            ]
        else:
            raise ValueError('invalid mode')
            
        oo = r + 1

        # On pairs:
        # The first component is used as is, the second is multiplied by the extra
        # variable. Note that index 0 corresponds to {1} and index oo to {0}.
        self.pairs = (
            [(oo,0)] +
            [(i,oo)  for i in xrange(1,r)] +
            [(i,i-1) for i in xrange(1,r+1)]
            )
        # Total number of pairs: 2 * r
        self.integrand = (
            (1,) + (2*r-1)*(0,),
            (self.d-r+1,) + (r-1) * (-1,) + r * (+1,)
            )
        self.datum = IgusaDatum(F + [LaurentIdeal(gens=[], RS=self.RS, ring=FractionField(actual_ring))]).simplify()
        return self.datum

    def padically_evaluate(self, shuffle=False):
        if self.root() is None:
            return SR(1)
        return ((1 - SR('q')**(-1))**(-1) * LocalZetaProcessor.padically_evaluate(self, shuffle=shuffle)).factor()

    def __repr__(self):
        if self.root() is not None:
            s = 'Ask processor:\n'
            s += 'Base ring: %s\n' % self.datum.ring
            s += 'R =\n' + str(self.R) + '\n'
            s += 'd = %d\n' % self.d
            s += 'e = %d\n' % self.e
            s += 'r = %d\n' % self.r
            s += 'Root:\n%s\n' % self.datum
            s += 'Pairs: ' + str(self.pairs) + '\n'
            s += 'Integrand: ' + str(self.integrand)
            return s
        else:
            return 'Trivial ask processor'



from sage.all import Infinity

from multiprocessing import cpu_count

# paths of programs. By default use programs in $PATH.
libcrunch = None
normaliz = "normaliz"
count = "count"

ncpus = cpu_count()
debug = False
disklist = False

symbolic = False

dict_polynomial = True
save_memory = False

_alt_ncpus = None
optimisation_level = 1
addmany_dispatcher = 'numerator'
symbolic_count_varieties = []

_simplify_bound = Infinity

# The plumber tries to prevent leaks. He does this by e.g. avoiding libsingular.
plumber = True



"""A stack that might live in a temporary file instead of system memory.
"""

from sage.all import *

import common

import contextlib, os, struct

from util import TemporaryDirectory

_INTSIZE = 8
_INTSYMB = '=q'

class DiskList(object):
    # Write operations are not thread safe, but reading is.
    # Exceptions during changes may leave the list in an inconsistent state.

    def __init__(self, filename):
        self._data  = filename
        self._index = filename + '.index'
        self._sizes = filename + '.sizes'

        # Make sure all files exist.
        with open(self._data, 'ab'), open(self._index, 'ab'), \
                open(self._sizes, 'ab'):
            pass

    def __len__(self):
        return os.path.getsize(self._index) / _INTSIZE

    def unlink(self):
        for f in [self._data, self._index, self._sizes]:
            os.remove(f)

    def append(self, x):
        s = dumps(x)

        with open(self._index, 'ab') as f:
            f.write(struct.pack(_INTSYMB, os.path.getsize(self._data)))
        with open(self._sizes, 'ab') as f:
            f.write(struct.pack(_INTSYMB, len(s)))
        with open(self._data, 'ab') as f:
            f.write(s)

    def extend(self, it):
        for x in it:
            self.append(x)

    def __getitem__(self, index):
        if index < 0:
            index = len(self) - index
        if index < 0 or index >= len(self):
                        raise IndexError('index out of range')

        # Get position
        with open(self._index, 'rb') as f:
            f.seek(_INTSIZE * index)
            pos, = struct.unpack(_INTSYMB, f.read(_INTSIZE))

        # Get size
        with open(self._sizes, 'rb') as f:
            f.seek(_INTSIZE * index)
            size, = struct.unpack(_INTSYMB, f.read(_INTSIZE))

        # Read data
        with open(self._data, 'rb') as f:
            f.seek(pos)
            return loads(f.read(size))

    def pop(self):
        length = len(self)
        if not length:
            raise IndexError('cannot pop an empty stack')

        # Read and delete the position of the last chunk.
        with open(self._index, 'r+b') as f:
            f.seek(_INTSIZE * (length-1))
            pos, = struct.unpack(_INTSYMB, f.read(_INTSIZE))
            f.seek(_INTSIZE * (length-1))
            f.truncate()

        # Read and delete the size of the last chunk.
        with open(self._sizes, 'r+b') as f:
            f.seek(_INTSIZE * (length-1))
            size, = struct.unpack(_INTSYMB, f.read(_INTSIZE))
            f.seek(_INTSIZE * (length-1))
            f.truncate()

        # Read and delete the chunk itself.
        with open(self._data, 'r+b') as f:
            f.seek(pos)
            top = f.read(size)
            f.seek(pos)
            f.truncate()
        return loads(top)

    def __iter__(self):
        for i in xrange(len(self)):
            yield self[i]

@contextlib.contextmanager
def TemporaryList(use_disk=False):
    if common.save_memory or use_disk:
        with TemporaryDirectory() as tmpdir:
            yield DiskList(os.path.join(tmpdir,'list'))
    else:
        yield []



"""
Some specialised functionality for cones, polyhedra, and polytopes.
"""

from sage.all import *
from sage.geometry.cone import ConvexRationalPolyhedralCone

import itertools
from collections import Iterable

import common

from util import monomial_exp, vertex_by_direction, cached_simple_method

import triangulate

def conify_polyhedron(P):
    """Compute the cone spanned by a polyhedron.
    """
    if isinstance(P, ConvexRationalPolyhedralCone):
        return P
    else:
        V = [vector(QQ,v) for v in P.vertices()]
        R = [vector(QQ,v) for v in P.rays()]
        Lplus = [vector(QQ,v) for v in P.lines()]
        Lminus = [-v for v in Lplus]
        return Cone(V+R+Lplus+Lminus, lattice=ZZ**P.ambient_dim())

def is_contained_in_dual(C, D):
    """Test if the cone C is contained in the dual of the cone D.
    """
    for v,w in itertools.product(C.rays(), D.rays()):
        if vector(QQ, v) * vector(QQ, w) < Rational(0):
            return False
    return True


def dual_cone_as_polyhedron(rays, strict=False):
    """Compute the polyhedron consisting of all y such that
    r * y >= 0 (or r * y > 0 if strict=True) for r in rays.
    """
    if len(rays) == 0:
        raise TypeError('Need at least one ray')
    c = Integer(-1) if strict else Integer(0)
    return Polyhedron(ieqs=[[c] + list(v) for v in rays], base_ring=QQ)

def inner_open_normal_fan(N):
    """
    Construct the relatively open, inner normal fan of a lattice
    polytope.

    We approximate 'x > 0' via 'x >= 1', see the part on 'models' of half-open
    cones in the second paper.
    """
    
    if N.is_empty():
        raise ValueError('need a non-empty polytope')

    V = [vector(QQ,v) for v in N.vertices()]

    if len(V) == 1:
        # Confusingly, given an empty list of (in)equalities, Sage
        # returns the empty polyhedron instead of the whole space.
        yield Polyhedron(eqns=[(N.ambient_dim()+1) * (0,)])
        return

    for face in N.face_lattice():
        if face.dim() == -1:
            continue
        W = [vector(QQ,w) for w in face.vertices()]

        eqns = [vector(QQ, [0] + list(W[0] - W[i])) for i in xrange(1,len(W))]
        idx = [i for i in xrange(len(V)) if not(V[i] in W)]
        ieqs = [vector(QQ, [-1] + list(V[i] - W[0])) for i in idx]
        yield Polyhedron(ieqs=ieqs, eqns=eqns)

    # This function will be called a lot so we cannot afford to cache
    # return values.
    N.face_lattice.clear_cache()

def linear_image_of_polyhedron(P, A):
    if A.nrows() != P.ambient_dim():
        raise ValueError('matrix does not act on ambient space of polyhedra')
    return Polyhedron(ambient_dim=P.ambient_dim(),
                      vertices=[vector(v)*A for v in P.vertices()],
                      rays=[vector(r)*A for r in P.rays()],
                      lines=[vector(l)*A for l in P.lines()])
    
def get_point_in_polyhedron(P):
    if len(P.vertices()) == 0:
        raise TypeError('What a strange polyhedron you gave me')
    return vector(QQ, P.vertices()[0])

def PositiveOrthant(d):
    if d == 0:
        return Polyhedron(ambient_dim=0,eqns=[],ieqs=[(0,)], base_ring=QQ)
    else:
        return Polyhedron(ieqs=block_matrix([[matrix(QQ,d,1),identity_matrix(QQ,d)]]))

def StrictlyPositiveOrthant(d):
    if d == 0:
        return Polyhedron(ambient_dim=0, vertices=[()], base_ring=QQ)
    else:
        return Polyhedron(ieqs=block_matrix([[matrix(QQ,d,1,[-1 for i in xrange(d)]),identity_matrix(QQ,d)]]))

def _mixed_volume_naive(gen):
    """
    Naive computation of the normalised mixed volume using Cox et al.,
    'Using Algebraic Geometry', Thm 7.4.12.
    """

    P = list(gen)
    n = len(P)

    if any(q.ambient_dim() != n for q in P):
        raise TypeError('Number of polytopes and ambient dimension do not match')
    res = 0
    for I in Subsets(range(n)):
        if not I:
            continue
        res += (-1)**len(I) * (sum(P[i] for i in I)).volume()

    return (-1)**n/Integer(factorial(n)) * res

def _mixed_volume_gfan(gen):
    """
    Use gfan to compute the mixed volume of a collection of polytopes.
    Just like Khovanskii, we use the normalised mixed volume which satisfies
    mixed_volume([P,...,P]) = volume(P).
    """

    P = list(gen)
    n = len(P)

    if any(Q.ambient_dim() != n for Q in P):
        raise TypeError('Number of polytopes and ambient dimension do not match')

    if n == 0:
        return 0
    elif n == 1:
        return P[0].volume()

    R = PolynomialRing(QQ, 'x', n)
    I = R.ideal([ sum(monomial_exp(R,e) for e in Q.vertices()) for Q in P ])
    return Integer(1)/Integer(factorial(n)) * I.groebner_fan().mixed_volume()

mixed_volume = _mixed_volume_gfan

def DirectProductOfPolyhedra(P, Q):
    m = P.ambient_dim()
    n = Q.ambient_dim()

    ieqs = []
    eqns = [(n+m+1)*(0,)] # ensures correctness when P == RR^m, Q==RR^n

    # [b | a] --> [b | a | 0]
    f = lambda v: v + n * [0]

    # [b | a] --> [b | 0 | a]
    g = lambda v: [v[0]] + m * [0] + v[1:]

    for i in P.inequalities():
        ieqs.append(f(list(i)))
    for i in Q.inequalities():
        ieqs.append(g(list(i)))

    for e in P.equations():
        eqns.append(f(list(e)))
    for e in Q.equations():
        eqns.append(g(list(e)))
    return Polyhedron(ieqs=ieqs, eqns=eqns, base_ring=QQ, ambient_dim=m+n)

class RationalSet(object):
    # Disjoint(!) unions of rational polyhedra.
    def __init__(self, arg, ambient_dim=None, check=False):
        try:
            polyhedra = list(Set(arg))
        except TypeError:
            polyhedra = [arg]

        if not polyhedra:
            if ambient_dim is None:
                raise ValueError('need to specify ambient dimension')
            self.ambient_dim = ambient_dim
            self.polyhedra = []
            self.cones = []
            return

        self.ambient_dim = polyhedra[0].ambient_dim() if ambient_dim is None else ambient_dim
        polyhedra = [P for P in polyhedra if not P.is_empty()]
        self.polyhedra = polyhedra
        self.cones = [conify_polyhedron(P) for P in polyhedra]

        if any(P.ambient_dim() != self.ambient_dim for P in polyhedra):
            raise ValueError('inconsistent ambient dimensions')

        if check and any(not (P & Q).is_empty() for P,Q in itertools.combinations(polyhedra, 2)):
            raise ValueError('polyhedra are not disjoint')

    def triangulate_max(self, dims=None):
        # Return a triangulation of the maximal-dimensional parts (or all
        # parts of specified dimensions) only.
        if dims is None:
            dims = [self.dim()]
        for C in self.cones:
            if C.dim() not in dims:
                continue
            if C.dim() == 0:
                raise RuntimError("what's a triangulation of the trivial cone?")
            for scone in triangulate.triangulate_cone(C):
                yield scone
    
    def topologise(self, Phi, dims=None):
        if self.dim() == 0:
            # NOTE: For the cone {0}, the generating function is just 1;
            # hence, Phi doesn't matter.
            yield triangulate.SURF(scalar=1, rays=[])
            return

        for scone in self.triangulate_max(dims=dims):
            yield triangulate._topologise_scone(scone, Phi)

    def generating_function(self, vars='x', base_list=None):
        if not self.polyhedra:
            raise NotImplementedError
        R = PolynomialRing(QQ, vars, self.ambient_dim)

        from smurf import SMURF
        if not common.count:
            raise RuntimeError("'count' needs to be present in order to compute generating functions")

        # f = SMURF.from_polyhedron if common.count else SMURF.from_half_open_cone
        f = SMURF.from_polyhedron

        return SMURF((sm for P in self.polyhedra for sm in f(P, R)),
                     base_list=base_list)

    def is_empty(self):
        return not self.polyhedra

    @cached_simple_method
    def dim(self):
        return -1 if not self.polyhedra else max(P.dim() for P in self.polyhedra)

    def intersection(self, other):
        if self.ambient_dim != other.ambient_dim:
            raise ValueError('ambient dimensions differ')
        return RationalSet([P & Q for P,Q in itertools.product(self.polyhedra, other.polyhedra)],
                           ambient_dim=self.ambient_dim)
    __and__ = intersection

    def __mul__(self, other):
        return RationalSet([DirectProductOfPolyhedra(P,Q) for P,Q in itertools.product(self.polyhedra, other.polyhedra)], ambient_dim=self.ambient_dim+other.ambient_dim)

    def dual_cone(self):
        return reduce(lambda C,D: C.intersection(D), (Cone(C.rays()).dual() for C in self.cones)) if self.cones else Cone(Polyhedron(ieqs=[(self.ambient_dim+1)*(0,)]))

    def is_contained_in_dual(self, alpha):
        return all(alpha*beta >= 0 for C in self.cones for beta in C.rays())
    
    def is_perpendicular(self, alpha):
        return all(alpha*beta == 0 for C in self.cones for beta in C.rays())

    def __repr__(self):
        s = 'Rational set consisting of %d polyhedra in RR^%d:\n' % (len(self.polyhedra), self.ambient_dim)
        for i, P in enumerate(self.polyhedra):
            s += '[%d]:\n\tdim = %d\n\t' % (i, P.dim())
            s += '\n\t'.join(str(h) for h in P.Hrep_generator()) + '\n'
        return s
